from __future__ import annotations

import contextlib
import functools
import os
import webbrowser
from typing import TYPE_CHECKING, Any, NoReturn

from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)
from public import public

import xorq.vendor.ibis.expr.operations as ops
from xorq.common.exceptions import TranslationError, XorqError
from xorq.common.utils.func_utils import return_constant
from xorq.vendor import ibis
from xorq.vendor.ibis.common.annotations import ValidationError
from xorq.vendor.ibis.common.grounds import Immutable
from xorq.vendor.ibis.common.patterns import Coercible, CoercionError
from xorq.vendor.ibis.common.typing import get_defining_scope
from xorq.vendor.ibis.config import options as opts
from xorq.vendor.ibis.expr.format import pretty
from xorq.vendor.ibis.util import deprecated, experimental


if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from rich.console import Console, RenderableType

    import xorq.vendor.ibis.expr.types as ir
    from xorq.vendor.ibis.backends import BaseBackend
    from xorq.vendor.ibis.expr.visualize import (
        EdgeAttributeGetter,
        NodeAttributeGetter,
    )


try:
    from rich.jupyter import JupyterMixin
except ImportError:

    class _FixedTextJupyterMixin:
        """No-op when rich is not installed."""
else:

    class _FixedTextJupyterMixin(JupyterMixin):
        """JupyterMixin adds a spurious newline to text, this fixes the issue."""

        def _repr_mimebundle_(self, *args, **kwargs):
            try:
                bundle = super()._repr_mimebundle_(*args, **kwargs)
            except Exception:  # noqa: BLE001
                return None
            else:
                bundle["text/plain"] = bundle["text/plain"].rstrip()
                return bundle


def _capture_rich_renderable(renderable: RenderableType) -> str:
    from rich.console import Console

    console = Console(force_terminal=False)
    with console.capture() as capture:
        console.print(renderable)
    return capture.get().rstrip()


@public
class Expr(Immutable, Coercible):
    """Base expression class."""

    __slots__ = ("_arg",)
    _arg: ops.Node

    def _noninteractive_repr(self) -> str:
        if ibis.options.repr.show_variables:
            scope = get_defining_scope(self, types=Expr)
        else:
            scope = None
        return pretty(self.op(), scope=scope)

    def __rich_console__(self, console: Console, options):
        from rich.text import Text

        if console.is_jupyter:
            # Rich infers a console width in jupyter notebooks, but since
            # notebooks can use horizontal scroll bars we don't want to apply a
            # limit here. Since rich requires an integer for max_width, we
            # choose an arbitrarily large integer bound. Note that we need to
            # handle this here rather than in `to_rich`, as this setting
            # also needs to be forwarded to `console.render`.
            options = options.update(max_width=1_000_000)
            console_width = None
        else:
            console_width = options.max_width

        try:
            if opts.interactive:
                from xorq.vendor.ibis.expr.types.pretty import to_rich

                rich_object = to_rich(self, console_width=console_width)
            else:
                rich_object = Text(self._noninteractive_repr())
        except TranslationError as e:
            lines = [
                "Translation to backend failed",
                f"Error message: {e!r}",
                "Expression repr follows:",
                self._noninteractive_repr(),
            ]
            return Text("\n".join(lines))
        return console.render(rich_object, options=options)

    def __repr__(self):
        from xorq.config import options

        if options.interactive:
            return _capture_rich_renderable(self)
        else:
            return self._noninteractive_repr()

    def __init__(self, arg: ops.Node) -> None:
        object.__setattr__(self, "_arg", arg)

    def __iter__(self) -> NoReturn:
        raise TypeError(f"{self.__class__.__name__!r} object is not iterable")

    @classmethod
    def __coerce__(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, ops.Node):
            return value.to_expr()
        else:
            raise CoercionError("Unable to coerce value to an expression")

    def __reduce__(self):
        return (self.__class__, (self._arg,))

    def __hash__(self):
        return hash((self.__class__, self._arg))

    def equals(self, other):
        """Return whether this expression is _structurally_ equivalent to `other`.

        If you want to produce an equality expression, use `==` syntax.

        Parameters
        ----------
        other
            Another expression

        Examples
        --------
        >>> import xorq.api as xo
        >>> t1 = xo.table(dict(a="int"), name="t")
        >>> t2 = xo.table(dict(a="int"), name="t")
        >>> t1.equals(t2)
        True
        >>> v = xo.table(dict(a="string"), name="v")
        >>> t1.equals(v)
        False
        """
        if not isinstance(other, Expr):
            raise TypeError(
                f"invalid equality comparison between Expr and {type(other)}"
            )
        return self._arg.equals(other._arg)

    def __bool__(self) -> bool:
        raise ValueError("The truth value of an Ibis expression is not defined")

    __nonzero__ = __bool__

    @deprecated(
        instead="remove any usage of `has_name`, since it is always `True`",
        as_of="9.4",
        removed_in="10.0",
    )
    def has_name(self):
        """Check whether this expression has an explicit name."""
        return hasattr(self._arg, "name")

    def get_name(self):
        """Return the name of this expression."""
        return self._arg.name

    def _repr_png_(self) -> bytes | None:
        if opts.interactive or not opts.graphviz_repr:
            return None
        try:
            import xorq.vendor.ibis.expr.visualize as viz
        except ImportError:
            return None
        else:
            # Something may go wrong, and we can't error in the notebook
            # so fallback to the default text representation.
            with contextlib.suppress(Exception):
                return viz.to_graph(self).pipe(format="png")

    def visualize(
        self,
        format: str = "svg",
        *,
        label_edges: bool = False,
        verbose: bool = False,
        node_attr: Mapping[str, str] | None = None,
        node_attr_getter: NodeAttributeGetter | None = None,
        edge_attr: Mapping[str, str] | None = None,
        edge_attr_getter: EdgeAttributeGetter | None = None,
    ) -> None:
        """Visualize an expression as a GraphViz graph in the browser.

        Parameters
        ----------
        format
            Image output format. These are specified by the `graphviz` Python
            library.
        label_edges
            Show operation input names as edge labels
        verbose
            Print the graphviz DOT code to stderr if [](`True`)
        node_attr
            Mapping of `(attribute, value)` pairs set for all nodes.
            Options are specified by the `graphviz` Python library.
        node_attr_getter
            Callback taking a node and returning a mapping of `(attribute, value)` pairs
            for that node. Options are specified by the `graphviz` Python library.
        edge_attr
            Mapping of `(attribute, value)` pairs set for all edges.
            Options are specified by the `graphviz` Python library.
        edge_attr_getter
            Callback taking two adjacent nodes and returning a mapping of `(attribute, value)` pairs
            for the edge between those nodes. Options are specified by the `graphviz` Python library.

        Examples
        --------
        Open the visualization of an expression in default browser:

        >>> import xorq.api as xo
        >>> import xorq.vendor.ibis.expr.operations as ops
        >>> left = ibis.table(dict(a="int64", b="string"), name="left")
        >>> right = ibis.table(dict(b="string", c="int64", d="string"), name="right")
        >>> expr = left.inner_join(right, "b").select(left.a, b=right.c, c=right.d)
        >>> expr.visualize(
        ...     format="svg",
        ...     label_edges=True,
        ...     node_attr={"fontname": "Roboto Mono", "fontsize": "10"},
        ...     node_attr_getter=lambda node: isinstance(node, ops.Field) and {"shape": "oval"},
        ...     edge_attr={"fontsize": "8"},
        ...     edge_attr_getter=lambda u, v: isinstance(u, ops.Field) and {"color": "red"},
        ... )  # quartodoc: +SKIP # doctest: +SKIP

        Raises
        ------
        ImportError
            If `graphviz` is not installed.
        """
        import xorq.vendor.ibis.expr.visualize as viz

        path = viz.draw(
            viz.to_graph(
                self,
                node_attr=node_attr,
                node_attr_getter=node_attr_getter,
                edge_attr=edge_attr,
                edge_attr_getter=edge_attr_getter,
                label_edges=label_edges,
            ),
            format=format,
            verbose=verbose,
        )
        webbrowser.open(f"file://{os.path.abspath(path)}")

    def pipe(self, f, *args: Any, **kwargs: Any) -> Expr:
        """Compose `f` with `self`.

        Parameters
        ----------
        f
            If the expression needs to be passed as anything other than the
            first argument to the function, pass a tuple with the argument
            name. For example, (f, 'data') if the function f expects a 'data'
            keyword
        args
            Positional arguments to `f`
        kwargs
            Keyword arguments to `f`

        Examples
        --------
        >>> import xorq.api as xo
        >>> xo.options.interactive = False
        >>> t = xo.table([("a", "int64"), ("b", "string")], name="t")
        >>> f = lambda a: (a + 1).name("a")
        >>> g = lambda a: (a * 2).name("a")
        >>> result1 = t.a.pipe(f).pipe(g)
        >>> result1
        r0 := UnboundTable: t
          a int64
          b string
        <BLANKLINE>
        a: r0.a + 1 * 2

        >>> result2 = g(f(t.a))  # equivalent to the above
        >>> result1.equals(result2)
        True

        Returns
        -------
        Expr
            Result type of passed function
        """
        if isinstance(f, tuple):
            f, data_keyword = f
            kwargs = kwargs.copy()
            kwargs[data_keyword] = self
            return f(*args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    def op(self) -> ops.Node:
        return self._arg

    def _find_backends(self) -> tuple[list[BaseBackend], bool]:
        """Return the possible backends for an expression.

        Returns
        -------
        list[BaseBackend]
            A list of the backends found.
        """

        import xorq.expr.relations as rel
        from xorq.common.utils.graph_utils import get_ordered_unique_sources

        node_types = (
            ops.UnboundTable,
            ops.DatabaseTable,
            ops.SQLQueryResult,
            rel.CachedNode,
            rel.Read,
        )
        found = self.op().find(node_types)
        bound = tuple(op for op in found if not isinstance(op, ops.UnboundTable))
        backends = tuple(get_ordered_unique_sources(bound))
        has_unbound = any(isinstance(op, ops.UnboundTable) for op in found)
        return backends, has_unbound

    def _find_backend(self, *, use_default=True) -> BaseBackend:
        """Find the backend attached to an expression.

        Parameters
        ----------
        use_default
            If [](`True`) and the default backend isn't set, initialize the
            default backend and use that. This should only be set to `True` for
            `.execute()`. For other contexts such as compilation, this option
            doesn't make sense so the default value is [](`False`).

        Returns
        -------
        BaseBackend
            A backend that is attached to the expression
        """
        from xorq.config import _backend_init

        backends, has_unbound = self._find_backends()

        match backends:
            case []:
                if has_unbound:
                    raise XorqError(
                        "Expression contains unbound tables and therefore cannot "
                        "be executed. Use `<backend>.execute(expr)` to execute "
                        "against an explicit backend, or rebuild the expression "
                        "using bound tables instead."
                    )
                elif use_default:
                    backend = _backend_init()
                    return backend
                else:
                    raise XorqError(
                        "Expression depends on no backends, and found no default"
                    )
            case [backend]:
                return backend
            case _:
                raise XorqError("Multiple backends found for this expression")

    def into_backend(self, con, name=None):
        """
        Converts the Expr to a table in the given backend `con` with an optional table name `name`.

        The table is backed by a PyArrow RecordBatchReader, the RecordBatchReader is teed
        so it can safely be reaused without spilling to disk.

        Parameters
        ----------
        con
            The backend where the table should be created
        name
            The name of the table

        Examples
        -------
        >>> import xorq.api as xo
        >>> from xorq.api import _
        >>> xo.options.interactive = True
        >>> ls_con = xo.connect()
        >>> pg_con = xo.postgres.connect_examples()
        >>> t = pg_con.table("batting").into_backend(ls_con, "ls_batting")
        >>> expr = (
        ...     t.join(t, "playerID")
        ...     .order_by("playerID", "yearID")
        ...     .limit(15)
        ...     .select(player_id="playerID", year_id="yearID_right")
        ... )
        >>> expr
        ┏━━━━━━━━━━━┳━━━━━━━━━┓
        ┃ player_id ┃ year_id ┃
        ┡━━━━━━━━━━━╇━━━━━━━━━┩
        │ string    │ int64   │
        ├───────────┼─────────┤
        │ aardsda01 │    2015 │
        │ aardsda01 │    2007 │
        │ aardsda01 │    2006 │
        │ aardsda01 │    2009 │
        │ aardsda01 │    2008 │
        │ aardsda01 │    2010 │
        │ aardsda01 │    2004 │
        │ aardsda01 │    2013 │
        │ aardsda01 │    2012 │
        │ aardsda01 │    2006 │
        │ …         │       … │
        └───────────┴─────────┘
        """

        from xorq.expr.relations import RemoteTable

        return RemoteTable.from_expr(con=con, expr=self, name=name).to_expr()

    def compile(
        self,
        limit: int | None = None,
        params: Mapping[ir.Value, Any] | None = None,
        pretty: bool = False,
    ):
        """Compile to an execution target.

        Parameters
        ----------
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value
        pretty
            In case of SQL backends, return a pretty formatted SQL query.
        """
        return self._find_backend().compile(
            self, limit=limit, params=params, pretty=pretty
        )

    def execute(self: ir.Expr, **kwargs: Any):
        """Execute an expression against its backend if one exists.

        Parameters
        ----------
        kwargs
            Keyword arguments

        Examples
        --------
        >>> import xorq.api as xo
        >>> t = xo.examples.penguins.fetch()
        >>> t.execute()
               species     island  bill_length_mm  ...  body_mass_g     sex  year
        0       Adelie  Torgersen            39.1  ...       3750.0    male  2007
        1       Adelie  Torgersen            39.5  ...       3800.0  female  2007
        2       Adelie  Torgersen            40.3  ...       3250.0  female  2007
        3       Adelie  Torgersen             NaN  ...          NaN    None  2007
        4       Adelie  Torgersen            36.7  ...       3450.0  female  2007
        ..         ...        ...             ...  ...          ...     ...   ...
        339  Chinstrap      Dream            55.8  ...       4000.0    male  2009
        340  Chinstrap      Dream            43.5  ...       3400.0  female  2009
        341  Chinstrap      Dream            49.6  ...       3775.0    male  2009
        342  Chinstrap      Dream            50.8  ...       4100.0    male  2009
        343  Chinstrap      Dream            50.2  ...       3775.0  female  2009
        [344 rows x 8 columns]

        Scalar parameters can be supplied dynamically during execution.
        >>> species = xo.param("string")
        >>> expr = t.filter(t.species == species).order_by(t.bill_length_mm)
        >>> expr.execute(limit=3, params={species: "Gentoo"})
          species  island  bill_length_mm  ...  body_mass_g     sex  year
        0  Gentoo  Biscoe            40.9  ...         4650  female  2007
        1  Gentoo  Biscoe            41.7  ...         4700  female  2009
        2  Gentoo  Biscoe            42.0  ...         4150  female  2007
        <BLANKLINE>
        [3 rows x 8 columns]
        """
        from xorq.expr.api import execute

        return execute(self, **kwargs)

    def to_pyarrow_batches(
        self: ir.Expr,
        *,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ):
        """Execute expression and return a RecordBatchReader.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        chunk_size
            Maximum number of rows in each returned record batch.
        kwargs
            Keyword arguments

        Returns
        -------
        results
            RecordBatchReader
        """
        from xorq.expr.api import to_pyarrow_batches

        return to_pyarrow_batches(self, chunk_size=chunk_size, **kwargs)

    def to_pyarrow(self: ir.Expr, **kwargs: Any):
        """Execute expression and return results in as a pyarrow table.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        kwargs
            Keyword arguments

        Returns
        -------
        Table
            A pyarrow table holding the results of the executed expression.
        """

        from xorq.expr.api import to_pyarrow

        return to_pyarrow(self, **kwargs)

    def to_parquet(
        self: ir.Expr,
        path: str | Path,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ):
        """Write the results of executing the given expression to a parquet file.

        This method is eager and will execute the associated expression
        immediately.

        See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetWriter.html for details.

        Parameters
        ----------
        path
            A string or Path where the Parquet file will be written.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to pyarrow.parquet.ParquetWriter

        Examples
        --------
        Write out an expression to a single parquet file.

        >>> import ibis
        >>> import tempfile
        >>> penguins = ibis.examples.penguins.fetch()
        >>> penguins.to_parquet(tempfile.mktemp())
        """
        from xorq.expr.api import to_parquet

        return to_parquet(self, path=path, params=params, **kwargs)

    @experimental
    def to_csv(
        self,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a CSV file.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to pyarrow.csv.CSVWriter

        https://arrow.apache.org/docs/python/generated/pyarrow.csv.CSVWriter.html
        """
        from xorq.expr.api import to_csv

        return to_csv(self, path=path, params=params, **kwargs)

    @experimental
    def to_json(
        self,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of `expr` to a NDJSON file.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        path
            The data source. A string or Path to the Delta Lake table.
        **kwargs
            Additional, backend-specific keyword arguments.

        https://github.com/ndjson/ndjson-spec
        """
        from xorq.expr.api import to_json

        return to_json(self, path=path, params=params)

    def unbind(self) -> ir.Table:
        """Return an expression built on `UnboundTable` instead of backend-specific objects."""
        from xorq.vendor.ibis.expr.rewrites import _, d, p

        rule = p.DatabaseTable >> d.UnboundTable(
            name=_.name, schema=_.schema, namespace=_.namespace
        )
        return self.op().replace(rule).to_expr()

    def as_table(self) -> ir.Table:
        """Convert an expression to a table."""
        raise NotImplementedError(
            f"{type(self)} expressions cannot be converted into tables"
        )

    def as_scalar(self) -> ir.Scalar:
        """Convert an expression to a scalar."""
        raise NotImplementedError(
            f"{type(self)} expressions cannot be converted into scalars"
        )

    @property
    def ls(self):
        return LETSQLAccessor(self.op())


@frozen
class LETSQLAccessor:
    op = field(validator=instance_of(ops.Node))
    node_types = (ops.DatabaseTable, ops.SQLQueryResult)

    @property
    def expr(self):
        return self.op.to_expr()

    @property
    def cached_nodes(self):
        from xorq.common.utils.graph_utils import walk_nodes
        from xorq.expr.relations import (
            CachedNode,
        )

        return walk_nodes((CachedNode,), self.expr)

    @property
    def tags(self):
        return self.get_tags()

    def get_tags(self, predicate=return_constant(True)):
        from xorq.common.utils.graph_utils import walk_nodes
        from xorq.expr.relations import Tag

        return tuple(
            node for node in walk_nodes((Tag,), self.expr) if predicate(node.metadata)
        )

    @property
    def cache(self):
        if self.is_cached:
            return self.op.cache
        else:
            return None

    @property
    def caches(self):
        return tuple(node.cache for node in self.cached_nodes)

    @property
    def backends(self):
        from xorq.common.utils.graph_utils import (
            find_all_sources,
        )

        return find_all_sources(self.expr)

    @property
    def is_multiengine(self):
        (_, *rest) = set(self.backends)
        return bool(rest)

    @property
    def dts(self):
        from xorq.common.utils.graph_utils import (
            walk_nodes,
        )
        from xorq.expr.relations import CachedNode, RemoteTable

        return tuple(
            el
            for el in walk_nodes(self.node_types, self.expr)
            if not isinstance(
                el,
                (
                    RemoteTable,
                    CachedNode,
                ),
            )
        )

    @property
    def is_cached(self):
        from xorq.expr.relations import (
            CachedNode,
        )

        return isinstance(self.op, CachedNode)

    @property
    def has_cached(self):
        return bool(self.cached_nodes)

    @property
    @functools.cache
    def untagged(self):
        from xorq.expr.api import (
            _remove_tag_nodes,
        )

        return _remove_tag_nodes(self.expr)

    @property
    def uncached(self):
        from xorq.expr.relations import CachedNode, RemoteTable, replace_cache_table

        if self.has_cached:
            op = self.expr.op()
            return op.replace(
                replace_cache_table, filter=(RemoteTable, CachedNode)
            ).to_expr()
        else:
            return self.expr

    @property
    def uncached_one(self):
        if self.is_cached:
            from xorq.expr.relations import RemoteTable

            parent = self.expr.op().parent
            if isinstance(parent.op(), RemoteTable):
                return parent.op().remote_expr
            else:
                return parent
        else:
            return self.expr

    @property
    def tokenized(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            patched_tokenize,
        )

        # NOTE: this should almost certainly not be functools.cache'd: it can obscure filesystem / source table changes within the same process run
        return patched_tokenize(self.expr)

    def get_cache_path(self):
        if self.is_cached and hasattr(self.cache.storage, "get_path"):
            cn = self.op
            return cn.cache.storage.get_path(cn.cache.calc_key(cn.parent))
        else:
            return None

    def get_cache_paths(self):
        if self.has_cached:
            return tuple(
                filter(
                    None, (op.to_expr().ls.get_cache_path() for op in self.cached_nodes)
                )
            )
        else:
            return None

    @property
    def cache_path(self):
        return self.get_cache_path()

    @property
    def cached_dt(self):
        if self.exists():
            return self.cache.get(self.uncached_one)
        else:
            return None

    def get_key(self):
        if self.is_cached:
            return self.cache.calc_key(self.uncached_one)
        else:
            return None

    def get_keys(self):
        if self.has_cached and self.cached_nodes[0].to_expr().ls.exists():
            # FIXME: yield cache with key
            return tuple(op.to_expr().ls.get_key() for op in self.cached_nodes)
        else:
            return None

    def exists(self):
        if self.is_cached:
            cn = self.op
            return cn.cache.exists(cn.parent)
        else:
            return None


def _binop(op_class: type[ops.Binary], left: ir.Value, right: ir.Value) -> ir.Value:
    """Try to construct a binary operation.

    Parameters
    ----------
    op_class
        The `ops.Binary` subclass for the operation
    left
        Left operand
    right
        Right operand

    Returns
    -------
    ir.Value
        A value expression

    Examples
    --------
    >>> import xorq.api as xo
    >>> import xorq.vendor.ibis.expr.operations as ops
    >>> expr = _binop(ops.TimeAdd, ibis.time("01:00"), ibis.interval(hours=1))
    >>> expr
    TimeAdd(datetime.time(1, 0), 1h): datetime.time(1, 0) + 1 h
    >>> _binop(ops.TimeAdd, 1, ibis.interval(hours=1))
    TimeAdd(datetime.time(0, 0, 1), 1h): datetime.time(0, 0, 1) + 1 h
    """
    try:
        node = op_class(left, right)
    except (ValidationError, NotImplementedError):
        return NotImplemented
    else:
        return node.to_expr()


def _is_null_literal(value: Any) -> bool:
    """Detect whether `value` will be treated by ibis as a null literal."""
    if value is None:
        return True
    if isinstance(value, Expr):
        op = value.op()
        return isinstance(op, ops.Literal) and op.value is None
    return False



import warnings

import dask
import toolz


try:
    import cityhash  # noqa: F401
except ImportError:
    warnings.warn(
        "cityhash is not installed, some functionality will not work",
        UserWarning,
        stacklevel=2,
    )

from xorq.common.utils.logging_utils import get_logger


logger = get_logger(__name__)


none_tokenized = dask.base.tokenize(None)


@toolz.curry
def hotfix(obj, attrname, target_tokenized, hotfix):
    return maybe_hotfix(obj, attrname, target_tokenized, hotfix, definitely=True)


@toolz.curry
def maybe_hotfix(obj, attrname, target_tokenized, hotfix, definitely=False):
    to_hotfix = getattr(obj, attrname, None)
    tokenized = dask.base.tokenize(to_hotfix)
    dct = {
        "obj.__name__": obj.__name__,
        "obj.__module__": obj.__module__,
        "attrname": attrname,
        "tokenized": tokenized,
        "target_tokenized": target_tokenized,
        "hotfix_tokenized": dask.base.tokenize(hotfix),
    }

    if definitely or (tokenized == target_tokenized):
        if not isinstance(hotfix, property):
            hotfix._original = to_hotfix
        else:
            if tokenized != none_tokenized:
                raise ValueError("Don't know how to retain _original")
        setattr(obj, attrname, hotfix)
        logger.info("hotfixing", **dct)
    else:
        logger.info("not hotfixing", **dct)
    return hotfix



import tempfile
from contextlib import contextmanager

from xorq.catalog.tar_utils import (
    extract_build_tgz_context,
    make_tgz_context,
)


@contextmanager
def build_expr_context(expr):
    from xorq.ibis_yaml.compiler import build_expr

    with tempfile.TemporaryDirectory() as td:
        build_dir = build_expr(expr, builds_dir=td)
        yield build_dir


@contextmanager
def build_expr_context_tgz(expr):
    with build_expr_context(expr) as build_dir:
        with make_tgz_context(build_dir) as tgz_path:
            yield tgz_path


def load_expr_from_tgz(tgz_path):
    from xorq.ibis_yaml.compiler import load_expr

    with extract_build_tgz_context(tgz_path) as build_dir:
        expr = load_expr(build_dir)
        return expr



from __future__ import annotations

import calendar
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import (
    FALSE,
    NULL,
    STAR,
    TRUE,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import MSSQLType
from xorq.vendor.ibis.backends.sql.dialects import MSSQL
from xorq.vendor.ibis.backends.sql.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_row_number,
    p,
    replace,
    split_select_distinct_with_order_by,
)
from xorq.vendor.ibis.common.deferred import var


if TYPE_CHECKING:
    from collections.abc import Mapping

    import xorq.vendor.ibis.expr.operations as ir

y = var("y")
start = var("start")
end = var("end")


# MS SQL facts that make using it a nightmare:
#
# * There is no boolean type
# * There are no boolean literals
# * But there's a numeric bit type whose domain is THE TWO VALUES 0 and 1 (and NULL of course), seriously?
# * Supported boolean expressions are =, <>, <, >, <=, >=, IS NULL,
#   IS NOT NULL, IN, NOT IN, EXISTS, BETWEEN, IS NOT DISTINCT FROM, IS DISTINCT FROM,
#   LIKE, NOT LIKE, CONTAINS (?), ALL, SOME, ANY
#   The return type of these is anyone's guess, but it's definitely NOT BOOLEAN
# * Boolean expressions CANNOT be used in a projection, i.e., SELECT x = 1 is not allowed
# * Boolean expressions MUST be used in a WHERE clause, i.e., SELECT * FROM t WHERE 1 is not allowed


@replace(p.WindowFunction(p.Reduction & ~p.ReductionVectorizedUDF, order_by=()))
def rewrite_rows_range_order_by_window(_, **kwargs):
    # MSSQL requires an order by in a window frame that has either ROWS or RANGE
    return _.copy(order_by=(_.func.arg,))


class MSSQLCompiler(SQLGlotCompiler):
    __slots__ = ()

    agg = AggGen(supports_order_by=True)

    dialect = MSSQL
    type_mapper = MSSQLType
    rewrites = (
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_row_number,
        rewrite_rows_range_order_by_window,
        *SQLGlotCompiler.rewrites,
    )
    post_rewrites = (split_select_distinct_with_order_by,)
    copy_func_args = True

    UNSUPPORTED_OPS = (
        ops.ApproxMedian,
        ops.ArgMax,
        ops.ArgMin,
        ops.Array,
        ops.ArrayDistinct,
        ops.ArrayFlatten,
        ops.ArrayMap,
        ops.ArraySort,
        ops.ArrayUnion,
        ops.BitAnd,
        ops.BitOr,
        ops.BitXor,
        ops.Covariance,
        ops.CountDistinctStar,
        ops.DateDiff,
        ops.IntervalAdd,
        ops.IntervalSubtract,
        ops.IntervalMultiply,
        ops.IntervalFloorDivide,
        ops.IsInf,
        ops.IsNan,
        ops.Levenshtein,
        ops.Map,
        ops.Median,
        ops.Mode,
        ops.NthValue,
        ops.RegexExtract,
        ops.RegexReplace,
        ops.RegexSearch,
        ops.RegexSplit,
        ops.RowID,
        ops.StringSplit,
        ops.StringToDate,
        ops.StringToTimestamp,
        ops.StructColumn,
        ops.TimestampDiff,
        ops.Unnest,
    )

    SIMPLE_OPS = {
        ops.Atan2: "atn2",
        ops.DateFromYMD: "datefromparts",
        ops.Hash: "checksum",
        ops.Ln: "log",
        ops.Log10: "log10",
        ops.Power: "power",
        ops.Repeat: "replicate",
        ops.Reverse: "reverse",
        ops.StringAscii: "ascii",
        ops.TimestampNow: "sysdatetime",
        ops.Min: "min",
        ops.Max: "max",
    }

    NAN = sg.func("double", sge.convert("NaN"))
    POS_INF = sg.func("double", sge.convert("Infinity"))
    NEG_INF = sg.func("double", sge.convert("-Infinity"))

    @staticmethod
    def _generate_groups(groups):
        return groups

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def to_sqlglot(
        self,
        expr: ir.Expr,
        *,
        limit: str | None = None,
        params: Mapping[ir.Expr, Any] | None = None,
    ):
        """Compile an Ibis expression to a sqlglot object."""
        from xorq.vendor import ibis

        table_expr = expr.as_table()
        conversions = {
            name: ibis.ifelse(table_expr[name], 1, 0).cast(dt.boolean)
            for name, typ in table_expr.schema().items()
            if typ.is_boolean()
        }

        if conversions:
            table_expr = table_expr.mutate(**conversions)
        return super().to_sqlglot(table_expr, limit=limit, params=params)

    def visit_RandomUUID(self, op, **_):
        return self.f.newid()

    def visit_StringLength(self, op, *, arg):
        """The MSSQL LEN function doesn't count trailing spaces.

        Also, DATALENGTH (the suggested alternative) counts bytes and thus its
        result depends on the string's encoding.

        https://learn.microsoft.com/en-us/sql/t-sql/functions/len-transact-sql?view=sql-server-ver16#remarks

        The solution is to add a character to the beginning and end of the
        string that are guaranteed to have one character in length and are not
        spaces, and then subtract 2 from the result of `LEN` of that input.

        Thanks to @arkanovicz for this glorious hack.
        """
        return sge.paren(self.f.len(self.f.concat("A", arg, "Z")) - 2, copy=False)

    def visit_Substring(self, op, *, arg, start, length):
        start += 1
        start = self.if_(start >= 1, start, start + self.f.length(arg))
        if length is None:
            # We don't need to worry about if start + length is greater than the
            # length of the string, MSSQL will just return the rest of the string
            length = self.f.length(arg)
        return self.f.substring(arg, start, length)

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        if where is not None:
            arg = self.if_(where, arg, NULL)

        out = self.f.group_concat(arg, sep)

        if order_by:
            out = sge.WithinGroup(this=out, expression=sge.Order(expressions=order_by))

        return out

    def visit_CountStar(self, op, *, arg, where):
        if where is not None:
            return self.f.sum(self.if_(where, 1, 0))
        return self.f.count(STAR)

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_ApproxQuantile(self, op, *, arg, quantile, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return sge.WithinGroup(
            this=self.f.approx_percentile_cont(quantile),
            expression=sge.Order(expressions=[sge.Ordered(this=arg, nulls_first=True)]),
        )

    def visit_DayOfWeekIndex(self, op, *, arg):
        return self.f.datepart(self.v.weekday, arg) - 1

    def visit_DayOfWeekName(self, op, *, arg):
        days = calendar.day_name
        return sge.Case(
            this=self.f.datepart(self.v.weekday, arg) - 1,
            ifs=list(map(self.if_, *zip(*enumerate(days)))),
        )

    def visit_DateTimestampTruncate(self, op, *, arg, unit):
        interval_units = {
            "us": "microsecond",
            "ms": "millisecond",
            "s": "second",
            "m": "minute",
            "h": "hour",
            "D": "day",
            "W": "week",
            "M": "month",
            "Q": "quarter",
            "Y": "year",
        }
        if (unit := interval_units.get(unit.short)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit!r}")

        return self.f.datetrunc(self.v[unit], arg, dialect=self.dialect)

    visit_DateTruncate = visit_TimestampTruncate = visit_DateTimestampTruncate

    def visit_Date(self, op, *, arg):
        return self.cast(arg, dt.date)

    def visit_DateTimeDelta(self, op, *, left, right, part):
        return self.f.datediff(
            sge.Var(this=part.this.upper()), right, left, dialect=self.dialect
        )

    visit_TimeDelta = visit_DateDelta = visit_TimestampDelta = visit_DateTimeDelta

    def visit_Xor(self, op, *, left, right):
        return sg.and_(sg.or_(left, right), sg.not_(sg.and_(left, right)))

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        interval_units = {
            "ms": "millisecond",
            "s": "second",
            "m": "minute",
            "h": "hour",
            "D": "day",
            "W": "week",
            "M": "month",
            "Q": "quarter",
            "Y": "year",
        }

        if not isinstance(op.interval, ops.Literal):
            raise com.UnsupportedOperationError(
                "Only literal interval values are supported with MS SQL timestamp bucketing"
            )

        if (unit := interval_units.get(op.interval.dtype.unit.short)) is None:
            raise com.UnsupportedOperationError(
                f"Unsupported bucket interval {op.interval!r}"
            )
        if offset is not None:
            raise com.UnsupportedOperationError(
                "Timestamp bucket with offset is not supported"
            )

        part = self.v[unit]
        origin = self.cast("1970-01-01", op.arg.dtype)

        return self.f.date_bucket(part, op.interval.value, arg, origin)

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.cast(
            self.f.datediff(self.v.s, "1970-01-01 00:00:00", arg, dialect=self.dialect),
            dt.int64,
        )

    def visit_ExtractTemporalComponent(self, op, *, arg):
        return self.f.datepart(self.v[type(op).__name__[len("Extract") :].lower()], arg)

    visit_ExtractYear = visit_ExtractMonth = visit_ExtractDay = (
        visit_ExtractDayOfYear
    ) = visit_ExtractHour = visit_ExtractMinute = visit_ExtractSecond = (
        visit_ExtractMillisecond
    ) = visit_ExtractMicrosecond = visit_ExtractTemporalComponent

    def visit_ExtractWeekOfYear(self, op, *, arg):
        return self.f.datepart(self.v.iso_week, arg)

    def visit_TimeFromHMS(self, op, *, hours, minutes, seconds):
        return self.f.timefromparts(hours, minutes, seconds, 0, 0)

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        return self.f.datetimefromparts(year, month, day, hours, minutes, seconds, 0)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if start is not None:
            return self.f.charindex(substr, arg, start)
        return self.f.charindex(substr, arg)

    def visit_Round(self, op, *, arg, digits):
        return self.f.round(arg, digits if digits is not None else 0)

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        unit = unit.short
        if unit == "s":
            return self.f.dateadd(self.v.s, arg, "1970-01-01 00:00:00")
        elif unit == "ms":
            return self.f.dateadd(self.v.s, arg / 1_000, "1970-01-01 00:00:00")
        raise com.UnsupportedOperationError(f"{unit!r} unit is not supported!")

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_decimal():
            return self.cast(str(value.normalize()), dtype)
        elif dtype.is_date():
            return self.f.datefromparts(value.year, value.month, value.day)
        elif dtype.is_timestamp():
            args = (
                value.year,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
                value.microsecond,
            )
            if dtype.timezone is not None:
                assert value.tzinfo is not None

                offset = value.strftime("%z")
                hour_offset = int(offset[:3])
                minute_offset = int(offset[-2:])
                return self.f.datetimeoffsetfromparts(
                    *args, hour_offset, minute_offset, 6
                )
            else:
                return self.f.datetime2fromparts(*args, 6)
        elif dtype.is_time():
            return self.f.timefromparts(
                value.hour, value.minute, value.second, value.microsecond, 0
            )
        elif dtype.is_uuid():
            return sge.Cast(
                this=sge.convert(str(value)),
                to=sge.DataType(this=sge.DataType.Type.UNIQUEIDENTIFIER),
            )
        elif dtype.is_binary():
            return self.f.convert(
                sge.DataType(this=sge.DataType.Type.VARBINARY),
                value.hex(),
                2,  # style, see https://learn.microsoft.com/en-us/sql/t-sql/functions/cast-and-convert-transact-sql?view=sql-server-ver16#binary-styles
            )
        elif dtype.is_array() or dtype.is_struct() or dtype.is_map():
            raise com.UnsupportedBackendType("MS SQL does not support complex types")

        return None

    def visit_Log2(self, op, *, arg):
        return self.f.log(arg, 2, dialect=self.dialect)

    def visit_Log(self, op, *, arg, base):
        if base is None:
            return self.f.log(arg, dialect=self.dialect)
        return self.f.log(arg, base, dialect=self.dialect)

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype

        if to.is_boolean():
            # no such thing as a boolean in MSSQL
            return arg
        elif from_.is_integer() and to.is_timestamp():
            return self.f.dateadd(self.v.s, arg, "1970-01-01 00:00:00")
        return super().visit_Cast(op, arg=arg, to=to)

    def visit_Sum(self, op, *, arg, where):
        if op.arg.dtype.is_boolean():
            arg = self.if_(arg, 1, 0)
        return self.agg.sum(arg, where=where)

    def visit_Mean(self, op, *, arg, where):
        if op.arg.dtype.is_boolean():
            arg = self.if_(arg, 1, 0)
        return self.agg.avg(arg, where=where)

    def visit_Not(self, op, *, arg):
        if isinstance(arg, sge.Boolean):
            return FALSE if arg == TRUE else TRUE
        elif isinstance(arg, (sge.Window, sge.Max, sge.Min)):
            # special case Window, Max, and Min.
            # These are used for NOT ANY or NOT ALL and friends.
            # We are working around MSSQL's rather unfriendly boolean handling rules
            # and because Max or Min don't return booleans, we have to handle the equality check
            # in a case statement instead.
            # e.g.
            # IFF(MAX(IFF(condition, 1, 0)) = 0, true_case, false_case)
            # is invalid
            # Needs to be
            # CASE WHEN MAX(IFF(condition, 1, 0)) = 0 THEN true_case ELSE false_case END
            return sge.Case(ifs=[self.if_(arg.eq(0), 1)], default=0)
        return self.if_(arg, 1, 0).eq(0)

    def visit_HashBytes(self, op, *, arg, how):
        if how in ("md5", "sha1"):
            return self.f.hashbytes(how, arg)
        elif how == "sha256":
            return self.f.hashbytes("sha2_256", arg)
        elif how == "sha512":
            return self.f.hashbytes("sha2_512", arg)
        else:
            raise NotImplementedError(how)

    def visit_HexDigest(self, op, *, arg, how):
        if how in ("md5", "sha1"):
            hashbinary = self.f.hashbytes(how, arg)
        elif how == "sha256":
            hashbinary = self.f.hashbytes("sha2_256", arg)
        elif how == "sha512":
            hashbinary = self.f.hashbytes("sha2_512", arg)
        else:
            raise NotImplementedError(how)

        # mssql uppercases the hexdigest which is inconsistent with several other
        # implementations and inconsistent with Python, so lowercase it.
        return self.f.lower(
            self.f.convert(
                sge.Literal(this="VARCHAR(MAX)", is_string=False), hashbinary, 2
            )
        )

    def visit_StringConcat(self, op, *, arg):
        any_args_null = (a.is_(NULL) for a in arg)
        return self.if_(sg.or_(*any_args_null), NULL, self.f.concat(*arg))

    def visit_Any(self, op, *, arg, where):
        arg = self.if_(arg, 1, 0)
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return sge.Max(this=arg)

    def visit_All(self, op, *, arg, where):
        arg = self.if_(arg, 1, 0)
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return sge.Min(this=arg)

    def visit_Select(
        self, op, *, parent, selections, predicates, qualified, sort_keys, distinct
    ):
        # if we've constructed a useless projection return the parent relation
        if not (selections or predicates or qualified or sort_keys or distinct):
            return parent

        result = parent

        if selections:
            result = sg.select(*self._cleanup_names(selections), copy=False).from_(
                result, copy=False
            )

        if predicates:
            result = result.where(*predicates, copy=True)

        if qualified:
            result = result.qualify(*qualified, copy=True)

        if sort_keys:
            result = result.order_by(*sort_keys, copy=False)

        if distinct:
            result = result.distinct()

        return result

    def visit_TimestampAdd(self, op, *, left, right):
        return self.f.dateadd(
            right.unit, self.cast(right.this, dt.int64), left, dialect=self.dialect
        )

    def visit_TimestampSub(self, op, *, left, right):
        return self.f.dateadd(
            right.unit, -self.cast(right.this, dt.int64), left, dialect=self.dialect
        )

    visit_DateAdd = visit_TimestampAdd
    visit_DateSub = visit_TimestampSub

    def visit_StartsWith(self, op, *, arg, start):
        return arg.like(self.f.concat(start, "%"))

    def visit_EndsWith(self, op, *, arg, end):
        return arg.like(self.f.concat("%", end))

    def visit_LPad(self, op, *, arg, length, pad):
        return self.if_(
            length <= self.f.length(arg),
            arg,
            self.f.left(
                self.f.concat(self.f.replicate(pad, length - self.f.length(arg)), arg),
                length,
            ),
        )

    def visit_RPad(self, op, *, arg, length, pad):
        return self.if_(
            length <= self.f.length(arg),
            arg,
            self.f.left(
                self.f.concat(arg, self.f.replicate(pad, length - self.f.length(arg))),
                length,
            ),
        )


compiler = MSSQLCompiler()



import re
import tarfile
from datetime import datetime
from functools import cache
from pathlib import Path

import yaml
from attr import field, frozen
from attr.validators import instance_of, optional
from textual import on, work
from textual.app import App, ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.screen import Screen
from textual.theme import Theme
from textual.widgets import (
    DataTable,
    Footer,
    Header,
    Static,
    TabbedContent,
    TabPane,
)
from toolz.curried import excepts as cexcepts

from xorq.common.utils.func_utils import return_constant


DEFAULT_REFRESH_INTERVAL = 10

XORQ_DARK = Theme(
    name="xorq-dark",
    primary="#C1F0FF",
    secondary="#4AA8EC",
    warning="#F5CA2C",
    error="#FF4757",
    success="#2BBE75",
    accent="#C1F0FF",
    foreground="#C1F0FF",
    background="#05181A",
    surface="#0a2a2e",
    panel="#0f3338",
    dark=True,
)

COLUMNS = ("KIND", "ALIAS", "HASH", "BACKEND", "OUTPUT", "CACHED", "ROOT TAG")

SCHEMA_PREVIEW_COLUMNS = ("NAME", "TYPE")

REVISION_COLUMNS = ("STATUS", "HASH", "COLUMNS", "CACHED", "DATE")

ALIAS_COLUMNS = ("ALIAS",)

GIT_LOG_COLUMNS = ("HASH", "DATE", "MESSAGE")


def maybe(default, exc=Exception):
    return cexcepts(exc, handler=return_constant(default))


def _format_cached(value: bool | None) -> str:
    match value:
        case True:
            return "●"
        case False:
            return "○"
        case _:
            return "—"


def _format_column_count(n: int | None) -> str:
    match n:
        case None:
            return "?"
        case int(n):
            return f"{n} cols"
        case _:
            return "?"


@frozen
class CatalogRowData:
    kind: str = field(default="expr", validator=instance_of(str))
    aliases: tuple[str, ...] = field(factory=tuple, validator=instance_of(tuple))
    hash: str = field(default="", validator=instance_of(str))
    backends: tuple[str, ...] = field(factory=tuple, validator=instance_of(tuple))
    column_count: int | None = field(default=None, validator=optional(instance_of(int)))
    cached: bool | None = field(default=None, validator=optional(instance_of(bool)))
    root_tag: str = field(default="", validator=instance_of(str))
    cached_expr: object = field(default=None, eq=False, hash=False, repr=False)

    @property
    @cache
    def aliases_display(self) -> str:
        return ", ".join(self.aliases) if self.aliases else ""

    @property
    @cache
    def backends_display(self) -> str:
        return ", ".join(sorted(set(self.backends))) if self.backends else ""

    @property
    @cache
    def output_display(self) -> str:
        return _format_column_count(self.column_count)

    @property
    @cache
    def cached_display(self) -> str:
        return _format_cached(self.cached)

    @property
    @cache
    def root_tag_display(self) -> str:
        return self.root_tag

    @property
    def sort_key(self) -> tuple[str, str]:
        return (self.aliases_display, self.hash)

    @property
    def row_key(self) -> str:
        return self.hash

    @property
    def row(self) -> tuple[str, ...]:
        return (
            self.kind,
            self.aliases_display,
            self.hash,
            self.backends_display,
            self.output_display,
            self.cached_display,
            self.root_tag_display,
        )


@frozen
class GitLogRowData:
    hash: str = field(default="", validator=instance_of(str))
    date: str = field(default="", validator=instance_of(str))
    message: str = field(default="", validator=instance_of(str))

    @property
    def row(self) -> tuple[str, ...]:
        return (self.hash, self.date, self.message)


@frozen
class ExploreData:
    hash: str = field(default="", validator=instance_of(str))
    alias: str = field(default="", validator=instance_of(str))
    schema_items: tuple[tuple[str, str], ...] = field(
        factory=tuple, validator=instance_of(tuple)
    )
    lineage_text: str = field(default="", validator=instance_of(str))
    is_cached: bool | None = field(default=None, validator=optional(instance_of(bool)))
    cache_path: str | None = field(default=None, validator=optional(instance_of(str)))
    metadata: tuple[tuple[str, str], ...] = field(
        factory=tuple, validator=instance_of(tuple)
    )
    has_alias: bool = field(default=False, validator=instance_of(bool))


@frozen
class RevisionRowData:
    hash: str = field(default="", validator=instance_of(str))
    column_count: int | None = field(default=None, validator=optional(instance_of(int)))
    cached: bool | None = field(default=None, validator=optional(instance_of(bool)))
    commit_date: str = field(default="", validator=instance_of(str))
    is_current: bool = field(default=False, validator=instance_of(bool))

    @property
    @cache
    def cached_display(self) -> str:
        return _format_cached(self.cached)

    @property
    @cache
    def status_display(self) -> str:
        return "CURRENT →" if self.is_current else ""

    @property
    @cache
    def columns_display(self) -> str:
        return _format_column_count(self.column_count)

    @property
    def row(self) -> tuple[str, ...]:
        return (
            self.status_display,
            self.hash,
            self.columns_display,
            self.cached_display,
            self.commit_date,
        )


def _check_cached(expr) -> bool:
    if not expr.ls.has_cached:
        return False
    return any(cn.to_expr().ls.exists() for cn in expr.ls.cached_nodes)


def _entry_info(entry) -> tuple[int, bool, str, object]:
    expr = entry.expr
    column_count = len(expr.columns)
    cached = _check_cached(expr)
    tags = expr.ls.tags
    root_tag = tags[0].tag if tags else ""
    return column_count, cached, root_tag or "", expr


def _extract_backends(entry) -> tuple[str, ...]:
    with tarfile.open(entry.catalog_path, "r:gz") as tf:
        f = tf.extractfile(f"{entry.name}/profiles.yaml")
        if f is None:
            return ()
        data = yaml.safe_load(f.read())
    if not isinstance(data, dict):
        return ()
    return tuple(
        pdata.get("con_name", "?") for pdata in data.values() if isinstance(pdata, dict)
    )


def _load_catalog_row(entry, aliases=()) -> CatalogRowData:
    try:
        column_count, cached, root_tag, expr = _entry_info(entry)
    except Exception:
        # expr deserialization can fail (e.g. missing module for pickled callable)
        column_count, cached, root_tag, expr = None, None, "", None
    return CatalogRowData(
        kind="expr",
        aliases=aliases,
        hash=entry.name,
        backends=_extract_backends(entry),
        column_count=column_count,
        cached=cached,
        root_tag=root_tag,
        cached_expr=expr,
    )


def _build_lineage_chain(expr) -> tuple[str, ...]:
    from xorq.common.utils.graph_utils import gen_children_of, to_node
    from xorq.common.utils.lineage_utils import format_node

    def _walk(node):
        yield format_node(node)
        match tuple(gen_children_of(node)):
            case (first, *_):
                yield from _walk(first)
            case _:
                pass

    return tuple(reversed(tuple(_walk(to_node(expr)))))


def _build_alias_multimap(
    catalog_aliases,
) -> dict[str, tuple[str, ...]]:
    from itertools import groupby
    from operator import attrgetter

    key = attrgetter("catalog_entry.name")
    sorted_aliases = sorted(catalog_aliases, key=key)
    return {
        name: tuple(sorted(ca.alias for ca in group))
        for name, group in groupby(sorted_aliases, key=key)
    }


def _build_git_log_rows(repo, max_count=100) -> tuple[GitLogRowData, ...]:
    return tuple(
        GitLogRowData(
            hash=commit.hexsha[:12],
            date=datetime.fromtimestamp(commit.committed_date).strftime(
                "%Y-%m-%d %H:%M"
            ),
            message=commit.message.strip().split("\n")[0],
        )
        for commit in repo.iter_commits(max_count=max_count)
    )


@maybe(default=None)
def maybe_expr(entry):
    return entry.expr


@maybe(default=())
def maybe_schema(expr) -> tuple[tuple[str, str], ...]:
    return tuple((name, str(dtype)) for name, dtype in expr.schema().items())


@maybe(default="(unavailable)")
def maybe_lineage(expr) -> str:
    chain = _build_lineage_chain(expr)
    return " → ".join(chain) if chain else "(empty)"


@maybe(default=None)
def maybe_cache_path(expr) -> str | None:
    paths = expr.ls.get_cache_paths()
    return str(paths[0]) if paths else None


def maybe_cache_info(expr) -> tuple[bool | None, str | None]:
    match expr:
        case None:
            return None, None
        case _ if not _check_cached(expr):
            return False, None
        case _:
            return True, maybe_cache_path(expr)


@maybe(default=())
def maybe_metadata(entry) -> tuple[tuple[str, str], ...]:
    if not entry.metadata_path.exists():
        return ()
    meta = yaml.safe_load(entry.metadata_path.read_text())
    match meta:
        case dict():
            return tuple((str(k), str(v)) for k, v in meta.items())
        case _:
            return ()


def _build_explore_data(
    entry, alias, known_cached=None, cached_expr=None
) -> ExploreData:
    expr = cached_expr if cached_expr is not None else maybe_expr(entry)
    computed_cached, cache_path = maybe_cache_info(expr)
    return ExploreData(
        hash=entry.name,
        alias=alias,
        schema_items=maybe_schema(expr),
        lineage_text=maybe_lineage(expr),
        is_cached=known_cached if known_cached is not None else computed_cached,
        cache_path=cache_path,
        metadata=maybe_metadata(entry),
        has_alias=bool(alias),
    )


# ---------------------------------------------------------------------------
# Screens
# ---------------------------------------------------------------------------


class CatalogScreen(Screen):
    BINDINGS = (
        ("q", "quit_app", "Quit"),
        ("ctrl+c", "quit_app", "Quit"),
        ("h", "scroll_left", "Left"),
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        ("l", "scroll_right", "Right"),
        ("enter", "explore", "Explore"),
        ("e", "explore", "Explore"),
        ("g", "toggle_git_log", "Git Log"),
    )

    def __init__(self, refresh_interval=DEFAULT_REFRESH_INTERVAL):
        super().__init__()
        self._refresh_interval = refresh_interval
        self._row_cache: dict[str, CatalogRowData] = {}
        self._git_log_visible = False
        self._git_log_loaded = False

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Vertical(id="catalog-panel"):
            yield DataTable(id="catalog-table")
        with Vertical(id="schema-panel"):
            yield DataTable(id="schema-preview-table")
        with Vertical(id="git-log-panel"):
            yield DataTable(id="git-log-table")
        yield Static("", id="status-bar")
        yield Footer()

    def on_mount(self) -> None:
        table = self.query_one("#catalog-table", DataTable)
        table.cursor_type = "row"
        table.zebra_stripes = True
        for col in COLUMNS:
            table.add_column(col, key=col)

        schema_table = self.query_one("#schema-preview-table", DataTable)
        schema_table.cursor_type = "row"
        schema_table.zebra_stripes = True
        for col in SCHEMA_PREVIEW_COLUMNS:
            schema_table.add_column(col, key=col)

        git_log_table = self.query_one("#git-log-table", DataTable)
        git_log_table.cursor_type = "row"
        git_log_table.zebra_stripes = True
        for col in GIT_LOG_COLUMNS:
            git_log_table.add_column(col, key=col)

        self.query_one("#catalog-panel").border_title = "Expressions"
        self.query_one("#schema-panel").border_title = "Schema"
        git_log_panel = self.query_one("#git-log-panel")
        git_log_panel.border_title = "Git Log"
        git_log_panel.display = False
        self.query_one("#status-bar", Static).update(" Loading catalog...")

        self.set_interval(self._refresh_interval, self._do_refresh)

    @on(DataTable.RowHighlighted, "#catalog-table")
    def _on_catalog_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        schema_table = self.query_one("#schema-preview-table", DataTable)
        schema_table.clear()
        if event.row_key is None:
            return
        row_data = self._row_cache.get(str(event.row_key.value))
        if row_data is None:
            return
        for name, dtype in maybe_schema(row_data.cached_expr):
            schema_table.add_row(name, dtype)

    @work(thread=True, exclusive=True)
    def _do_refresh(self) -> None:
        catalog = self.app._catalog
        if catalog is None:
            return
        repo_path = catalog.repo.working_dir
        expected_keys = frozenset(catalog.list())

        # build alias multimap: hash -> sorted tuple of aliases
        alias_multimap = _build_alias_multimap(catalog.catalog_aliases)

        # preserve insertion order from _row_cache (dict is ordered in Python 3.7+)
        cached_rows = tuple(
            self._row_cache[k] for k in self._row_cache if k in expected_keys
        )
        new_keys = expected_keys - self._row_cache.keys()
        removed = self._row_cache.keys() - expected_keys

        # evict removed entries
        self._row_cache = {
            k: v for k, v in self._row_cache.items() if k in expected_keys
        }

        match (bool(removed), bool(cached_rows)):
            case (True, _) | (_, False):
                # re-render when rows were removed or on first refresh
                self.app.call_from_thread(self._render_refresh, repo_path, cached_rows)
            case _:
                pass

        # load new entries incrementally (expensive I/O, off the main thread)
        for entry_hash in new_keys:
            entry = catalog.get_catalog_entry(entry_hash)
            aliases = alias_multimap.get(entry_hash, ())
            row_data = _load_catalog_row(entry, aliases)
            self._row_cache[row_data.row_key] = row_data
            self.app.call_from_thread(self._render_catalog_row, row_data)

        if self._git_log_visible:
            git_rows = _build_git_log_rows(catalog.repo)
            self.app.call_from_thread(self._render_git_log, git_rows)

        stamp = datetime.now().strftime("%H:%M:%S")
        self.app.call_from_thread(self._render_status, stamp, repo_path)

    def _render_refresh(self, repo_path, cached_rows) -> None:
        with self.app.batch_update():
            catalog_name = Path(repo_path).name
            self.query_one(
                "#catalog-panel"
            ).border_title = f"Expressions — {catalog_name}"

            table = self.query_one("#catalog-table", DataTable)
            saved_cursor = table.cursor_row
            table.clear()
            for row_data in cached_rows:
                table.add_row(*row_data.row, key=row_data.row_key)
            count = table.row_count
            if saved_cursor is not None and count > 0:
                table.move_cursor(row=min(saved_cursor, count - 1))

    def _render_catalog_row(self, row_data) -> None:
        with self.app.batch_update():
            self.query_one("#catalog-table", DataTable).add_row(
                *row_data.row, key=row_data.row_key
            )

    def _render_status(self, stamp, repo_path) -> None:
        count = self.query_one("#catalog-table", DataTable).row_count
        self.query_one("#status-bar", Static).update(
            f" {count} entries | {repo_path} | refreshed {stamp}"
        )

    def action_toggle_git_log(self) -> None:
        self._git_log_visible = not self._git_log_visible
        self.query_one("#git-log-panel").display = self._git_log_visible
        if self._git_log_visible and not self._git_log_loaded:
            self._load_git_log()

    @work(thread=True)
    def _load_git_log(self) -> None:
        catalog = self.app._catalog
        if catalog is None:
            return
        rows = _build_git_log_rows(catalog.repo)
        self._git_log_loaded = True
        self.app.call_from_thread(self._render_git_log, rows)

    def _render_git_log(self, rows) -> None:
        with self.app.batch_update():
            table = self.query_one("#git-log-table", DataTable)
            table.clear()
            for i, row_data in enumerate(rows):
                table.add_row(*row_data.row, key=str(i))

    def _focused_table(self) -> DataTable:
        focused = self.app.focused
        if isinstance(focused, DataTable):
            return focused
        return self.query_one("#catalog-table", DataTable)

    def action_scroll_left(self) -> None:
        self._focused_table().action_scroll_left()

    def action_cursor_down(self) -> None:
        self._focused_table().action_cursor_down()

    def action_cursor_up(self) -> None:
        self._focused_table().action_cursor_up()

    def action_scroll_right(self) -> None:
        self._focused_table().action_scroll_right()

    def action_quit_app(self) -> None:
        self.app.exit()

    def action_explore(self) -> None:
        table = self.query_one("#catalog-table", DataTable)
        if table.row_count == 0:
            return
        row_key, _ = table.coordinate_to_cell_key(table.cursor_coordinate)
        entry_hash = str(row_key.value)
        catalog = self.app._catalog
        try:
            entry = catalog.get_catalog_entry(entry_hash)
        except (AssertionError, KeyError):
            self.notify("Entry not found", severity="error")
            return
        row_data = self._row_cache.get(entry_hash)
        match row_data:
            case CatalogRowData(aliases=(first_alias, *_)):
                catalog_alias = next(
                    (ca for ca in catalog.catalog_aliases if ca.alias == first_alias),
                    None,
                )
            case _:
                first_alias = ""
                catalog_alias = None
        self.app.push_screen(
            ExploreScreen(
                entry, first_alias, catalog_alias=catalog_alias, row_data=row_data
            )
        )


class ExploreScreen(Screen):
    BINDINGS = (
        ("q", "go_back", "Back"),
        ("escape", "go_back", "Back"),
        ("ctrl+c", "quit_app", "Quit"),
        ("1", "tab_revisions", "Revisions"),
        ("2", "tab_schema", "Schema"),
        ("3", "tab_data", "Data"),
        ("4", "tab_info", "Info"),
        ("5", "tab_profiles", "Profiles"),
        ("6", "tab_aliases", "Aliases"),
        ("h", "scroll_left", "Left"),
        ("j", "cursor_down", "Down"),
        ("k", "cursor_up", "Up"),
        ("l", "scroll_right", "Right"),
        ("enter", "select_row", "Select"),
    )

    def __init__(self, entry, alias, catalog_alias=None, row_data=None):
        super().__init__()
        self._entry = entry
        self._alias = alias
        self._catalog_alias = catalog_alias
        self._row_data = row_data
        self._explore_data = None
        self._data_loaded = False
        self._revisions_loaded = False
        self._revisions = ()

    def compose(self) -> ComposeResult:
        alias_part = f" ({self._alias})" if self._alias else ""
        yield Header(show_clock=True)
        yield Static(
            f" {self._entry.name[:12]}{alias_part}",
            id="breadcrumb",
        )
        with TabbedContent(id="explore-tabs"):
            with TabPane("Revisions", id="pane-revisions", disabled=True):
                yield Static("Loading...", id="revisions-status")
                yield DataTable(id="revisions-table")
            with TabPane("Schema", id="pane-schema"):
                yield DataTable(id="schema-table")
            with TabPane("Data", id="pane-data", disabled=True):
                yield Static("Loading...", id="data-status")
                yield DataTable(id="data-table")
            with TabPane("Info", id="pane-info"):
                with VerticalScroll(id="info-scroll"):
                    with Vertical(id="lineage-section", classes="info-section"):
                        yield Static("", id="lineage-content")
                    with Vertical(id="cache-section", classes="info-section"):
                        yield Static("", id="cache-content")
                    with Vertical(id="metadata-section", classes="info-section"):
                        yield Static("", id="metadata-content")
            with TabPane("Profiles", id="pane-profiles"):
                yield DataTable(id="profiles-table")
            with TabPane("Aliases", id="pane-aliases"):
                yield DataTable(id="aliases-table")
        yield Footer()

    def on_mount(self) -> None:
        schema_table = self.query_one("#schema-table", DataTable)
        schema_table.cursor_type = "row"
        schema_table.zebra_stripes = True
        schema_table.add_column("NAME", key="name")
        schema_table.add_column("TYPE", key="type")

        aliases_table = self.query_one("#aliases-table", DataTable)
        aliases_table.cursor_type = "row"
        aliases_table.zebra_stripes = True
        for col in ALIAS_COLUMNS:
            aliases_table.add_column(col, key=col)

        data_table = self.query_one("#data-table", DataTable)
        data_table.cursor_type = "none"
        data_table.zebra_stripes = True
        data_table.loading = True

        rev_table = self.query_one("#revisions-table", DataTable)
        rev_table.cursor_type = "row"
        rev_table.zebra_stripes = True
        for col in REVISION_COLUMNS:
            rev_table.add_column(col, key=col)

        profiles_table = self.query_one("#profiles-table", DataTable)
        profiles_table.cursor_type = "row"
        profiles_table.zebra_stripes = True
        profiles_table.add_column("NAME", key="name")
        profiles_table.add_column("BACKEND", key="backend")
        profiles_table.add_column("PARAMETERS", key="params")
        profiles_table.add_column("ENV VARS", key="env_vars")
        self._load_profiles()

        self.query_one("#lineage-section").border_title = "Lineage"
        self.query_one("#cache-section").border_title = "Cache"
        self.query_one("#metadata-section").border_title = "Metadata"
        self.query_one("#metadata-section").display = False

        self._load_explore_data()

    @work(thread=True)
    def _load_profiles(self) -> None:
        env_re = re.compile(r"^\$\{(.+)\}$|^\$(.+)$")

        def _extract_env_vars(kwargs):
            return tuple(
                m.group(1) or m.group(2)
                for v in kwargs.values()
                if isinstance(v, str) and (m := env_re.match(v))
            )

        try:
            with tarfile.open(self._entry.catalog_path, "r:gz") as tf:
                f = tf.extractfile(f"{self._entry.name}/profiles.yaml")
                if f is None:
                    return
                data = yaml.safe_load(f.read())
            if not isinstance(data, dict):
                return
            rows = tuple(
                (
                    pname,
                    pdata.get("con_name", "?"),
                    ", ".join(
                        f"{k}={v}"
                        for k, v in (pdata.get("kwargs_tuple") or {}).items()
                        if v is not None
                    ),
                    ", ".join(_extract_env_vars(pdata.get("kwargs_tuple") or {})),
                )
                for pname, pdata in data.items()
            )
            self.app.call_from_thread(self._render_profiles, rows)
        except Exception:
            pass

    def _render_profiles(self, rows) -> None:
        table = self.query_one("#profiles-table", DataTable)
        for i, (name, backend, params, env_vars) in enumerate(rows):
            table.add_row(name, backend, params, env_vars, key=str(i))

    @work(thread=True)
    def _load_explore_data(self) -> None:
        known_cached = self._row_data.cached if self._row_data is not None else None
        cached_expr = self._row_data.cached_expr if self._row_data is not None else None
        data = _build_explore_data(
            self._entry,
            self._alias,
            known_cached=known_cached,
            cached_expr=cached_expr,
        )
        self.app.call_from_thread(self._render_explore, data)

    def _render_explore(self, data) -> None:
        self._explore_data = data

        schema_table = self.query_one("#schema-table", DataTable)
        if data.schema_items:
            for name, dtype in data.schema_items:
                schema_table.add_row(name, dtype)
        else:
            schema_table.add_row("(unavailable)", "")

        self.query_one("#lineage-content", Static).update(data.lineage_text)

        match (data.is_cached, data.cache_path):
            case (True, str() as path):
                cache_text = f"● cached\n  Path: {path}"
            case (True, _):
                cache_text = "● cached"
            case (False, _):
                cache_text = "○ uncached"
            case _:
                cache_text = "— unknown"
        self.query_one("#cache-content", Static).update(cache_text)

        if data.metadata:
            self.query_one("#metadata-section").display = True
            metadata_text = "\n".join(f"{k}: {v}" for k, v in data.metadata)
            self.query_one("#metadata-content", Static).update(metadata_text)

        aliases_table = self.query_one("#aliases-table", DataTable)
        aliases_table.clear()
        if self._row_data is not None:
            for i, alias_name in enumerate(self._row_data.aliases):
                aliases_table.add_row(alias_name, key=str(i))

        if data.is_cached is True:
            self.query_one("#pane-data", TabPane).disabled = False
            self.query_one("#data-status", Static).update(
                " Data preview (select tab to load)"
            )
        else:
            self.query_one("#data-status", Static).update(
                " uncached — run to materialize"
            )

        if data.has_alias:
            self.query_one("#pane-revisions", TabPane).disabled = False
            self._revisions_loaded = True
            self._load_revisions()
        else:
            self.query_one("#revisions-status", Static).update(
                " No alias — revisions unavailable"
            )

    @on(TabbedContent.TabActivated)
    def _on_tab_activated(self, event: TabbedContent.TabActivated) -> None:
        match event.pane.id:
            case "pane-data":
                if (
                    not self._data_loaded
                    and self._explore_data
                    and self._explore_data.is_cached is True
                ):
                    self._data_loaded = True
                    self._load_data_preview()
            case "pane-revisions":
                if (
                    not self._revisions_loaded
                    and self._explore_data
                    and self._explore_data.has_alias
                ):
                    self._revisions_loaded = True
                    self._load_revisions()
            case _:
                pass

    @work(thread=True)
    def _load_data_preview(self) -> None:
        try:
            expr = (
                self._row_data.cached_expr
                if self._row_data is not None and self._row_data.cached_expr is not None
                else self._entry.expr
            )
            df = self._execute_preview(expr)
            columns = tuple(str(c) for c in df.columns)
            rows = tuple(
                tuple(str(round(v, 2)) if isinstance(v, float) else str(v) for v in row)
                for row in df.itertuples(index=False)
            )
            total_rows = len(df)
            self.app.call_from_thread(self._render_data, columns, rows, total_rows)
        except Exception as e:
            self.app.call_from_thread(self._render_data_error, str(e))

    def _execute_preview(self, expr):
        try:
            return expr.head(50).execute()
        except Exception:
            # connection may be stale; reload from catalog entry and retry
            fresh_expr = self._entry.expr
            return fresh_expr.head(50).execute()

    def _render_data(self, columns, rows, total_rows) -> None:
        self.query_one("#data-status", Static).update(
            f" Data Preview — {total_rows} rows shown (max 50)"
        )
        data_table = self.query_one("#data-table", DataTable)
        data_table.loading = False
        for col in columns:
            data_table.add_column(col, key=col)
        for i, row in enumerate(rows):
            data_table.add_row(*row, key=str(i))
        data_table.cursor_type = "row"

    def _render_data_error(self, message) -> None:
        self.query_one("#data-status", Static).update(f" Error loading data: {message}")
        self.query_one("#data-table", DataTable).loading = False

    @work(thread=True)
    def _load_revisions(self) -> None:
        if self._catalog_alias is None:
            self.app.call_from_thread(
                self._render_revisions_error, "Alias object not available"
            )
            return
        try:
            raw_revisions = self._catalog_alias.list_revisions()
        except Exception as e:
            self.app.call_from_thread(
                self._render_revisions_error,
                f"Failed to load revisions: {e}",
            )
            return

        def _revision_pair(i, rev_entry, commit):
            try:
                exists = rev_entry.exists()
            except Exception:
                exists = False
            col_count, cached, _, _ = (
                _entry_info(rev_entry) if exists else (None, None, "", None)
            )
            row = RevisionRowData(
                hash=rev_entry.name,
                column_count=col_count,
                cached=cached,
                commit_date=datetime.fromtimestamp(commit.committed_date).strftime(
                    "%Y-%m-%d %H:%M"
                ),
                is_current=(i == 0),
            )
            return row, (rev_entry, commit, exists)

        pairs = tuple(
            _revision_pair(i, rev_entry, commit)
            for i, (rev_entry, commit) in enumerate(raw_revisions)
        )
        revision_rows = tuple(row for row, _ in pairs)
        valid_revisions = tuple(info for _, info in pairs)
        self.app.call_from_thread(
            self._render_revisions,
            revision_rows,
            valid_revisions,
        )

    def _render_revisions(self, revision_rows, valid_revisions) -> None:
        self._revisions = valid_revisions
        alias_name = (
            (self._alias or self._explore_data.alias) if self._explore_data else "?"
        )
        self.query_one("#revisions-status", Static).update(
            f" Revisions for '{alias_name}' — {len(revision_rows)} revisions"
        )
        rev_table = self.query_one("#revisions-table", DataTable)
        for i, row_data in enumerate(revision_rows):
            rev_table.add_row(*row_data.row, key=str(i))

    def _render_revisions_error(self, message) -> None:
        self.query_one("#revisions-status", Static).update(f" Error: {message}")

    def _active_table(self) -> DataTable | None:
        tabs = self.query_one("#explore-tabs", TabbedContent)
        match tabs.active:
            case "pane-schema":
                return self.query_one("#schema-table", DataTable)
            case "pane-aliases":
                return self.query_one("#aliases-table", DataTable)
            case "pane-data":
                return self.query_one("#data-table", DataTable)
            case "pane-revisions":
                return self.query_one("#revisions-table", DataTable)
            case "pane-profiles":
                return self.query_one("#profiles-table", DataTable)
            case _:
                return None

    def action_go_back(self) -> None:
        self.dismiss()

    def action_quit_app(self) -> None:
        self.app.exit()

    def action_scroll_left(self) -> None:
        table = self._active_table()
        if table is not None:
            table.action_scroll_left()

    def action_cursor_down(self) -> None:
        table = self._active_table()
        if table is not None:
            table.action_cursor_down()
        else:
            self.query_one("#info-scroll", VerticalScroll).scroll_down()

    def action_cursor_up(self) -> None:
        table = self._active_table()
        if table is not None:
            table.action_cursor_up()
        else:
            self.query_one("#info-scroll", VerticalScroll).scroll_up()

    def action_scroll_right(self) -> None:
        table = self._active_table()
        if table is not None:
            table.action_scroll_right()

    def action_tab_schema(self) -> None:
        self.query_one("#explore-tabs", TabbedContent).active = "pane-schema"

    def action_tab_aliases(self) -> None:
        self.query_one("#explore-tabs", TabbedContent).active = "pane-aliases"

    def action_tab_data(self) -> None:
        pane = self.query_one("#pane-data", TabPane)
        if pane.disabled:
            self.notify("Data tab unavailable (not cached)", severity="warning")
            return
        self.query_one("#explore-tabs", TabbedContent).active = "pane-data"

    def action_tab_revisions(self) -> None:
        pane = self.query_one("#pane-revisions", TabPane)
        if pane.disabled:
            self.notify("Revisions tab unavailable (no alias)", severity="warning")
            return
        self.query_one("#explore-tabs", TabbedContent).active = "pane-revisions"

    def action_tab_info(self) -> None:
        self.query_one("#explore-tabs", TabbedContent).active = "pane-info"

    def action_tab_profiles(self) -> None:
        self.query_one("#explore-tabs", TabbedContent).active = "pane-profiles"

    def action_select_row(self) -> None:
        tabs = self.query_one("#explore-tabs", TabbedContent)
        if tabs.active != "pane-revisions":
            return
        rev_table = self.query_one("#revisions-table", DataTable)
        if rev_table.row_count == 0:
            return
        row_key, _ = rev_table.coordinate_to_cell_key(rev_table.cursor_coordinate)
        idx = int(row_key.value)
        if idx >= len(self._revisions):
            return
        rev_entry, _, exists = self._revisions[idx]
        if not exists:
            self.notify("Entry no longer exists on disk", severity="warning")
            return
        explore_screens = tuple(
            s for s in self.app.screen_stack if isinstance(s, ExploreScreen)
        )
        if len(explore_screens) >= 2:
            self.notify("Maximum drill-down depth reached", severity="warning")
            return
        self.app.push_screen(ExploreScreen(rev_entry, ""))


class CatalogTUI(App):
    TITLE = "xorq catalog"
    CSS = """
    #catalog-panel {
        height: 2fr;
        border: solid $primary;
        background: $surface;
    }
    #catalog-table {
        height: 1fr;
    }
    #schema-panel {
        height: 1fr;
        border: solid $primary;
    }
    #schema-preview-table {
        height: 1fr;
    }
    #git-log-panel {
        height: 1fr;
        border: solid $primary;
    }
    #git-log-table {
        height: 1fr;
    }
    #status-bar {
        dock: bottom;
        height: 1;
        padding: 0 2;
    }
    #breadcrumb {
        height: 1;
        text-style: bold;
        padding: 0 2;
    }
    #explore-tabs {
        height: 1fr;
    }
    #schema-table {
        height: 1fr;
    }
    #data-status {
        height: 1;
        padding: 0 2;
    }
    #data-table {
        height: 1fr;
    }
    #revisions-status {
        height: 1;
        padding: 0 2;
    }
    #revisions-table {
        height: 1fr;
    }
    #info-scroll {
        height: 1fr;
    }
    .info-section {
        height: auto;
        margin: 0 1 1 1;
        padding: 1 2;
    }
    .info-section Static {
        height: auto;
    }
    #aliases-table {
        height: 1fr;
    }
    #profiles-table {
        height: 1fr;
    }
    """

    def __init__(self, make_catalog, refresh_interval=DEFAULT_REFRESH_INTERVAL):
        super().__init__()
        self._catalog = None
        self._make_catalog = make_catalog
        self._refresh_interval = refresh_interval
        self.register_theme(XORQ_DARK)
        self.theme = "xorq-dark"

    def on_mount(self) -> None:
        self.push_screen(CatalogScreen(refresh_interval=self._refresh_interval))
        self._load_catalog()

    @work(thread=True)
    def _load_catalog(self) -> None:
        catalog = self._make_catalog()
        self.app.call_from_thread(self._set_catalog, catalog)

    def _set_catalog(self, catalog) -> None:
        self._catalog = catalog
        self.screen._do_refresh()



# https://stackoverflow.com/a/12141386
from ctypes import (
    POINTER,
    Structure,
    c_size_t,
    c_void_p,
    cast,
    py_object,
)


PyObject_HEAD = [
    ("ob_refcnt", c_size_t),
    ("ob_type", c_void_p),
]


def make_class(*fields):
    _fields_ = PyObject_HEAD + [(field, c_void_p) for field in fields]
    return type("ctypes-hack", (Structure,), {"_fields_": _fields_})


# FIXME: define singledispatch and register common types
def get_ctypes_field(fields, field, obj):
    assert field in fields
    c_methcallobj = cast(c_void_p(id(obj)), POINTER(make_class(*fields))).contents

    return cast(getattr(c_methcallobj, field), py_object).value



import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_string_concat(compiler):
    s1 = ibis.literal("hello")
    s2 = ibis.literal("world")
    expr = s1 + s2
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "StringConcat"
    assert expression["arg"]["values"][0]["value"] == "hello"
    assert expression["arg"]["values"][1]["value"] == "world"
    assert dtype_yaml == {
        "op": "DataType",
        "type": "String",
        "nullable": {"op": "bool", "value": True},
    }


def test_string_upper_lower(compiler):
    s = ibis.literal("Hello")
    upper_expr = s.upper()
    lower_expr = s.lower()

    upper_yaml = compiler.to_yaml(upper_expr)["expression"]
    assert upper_yaml["op"] == "Uppercase"
    assert upper_yaml["arg"]["value"] == "Hello"

    lower_yaml = compiler.to_yaml(lower_expr)["expression"]
    assert lower_yaml["op"] == "Lowercase"
    assert lower_yaml["arg"]["value"] == "Hello"


def test_string_to_date(compiler):
    value = "20170206"
    format_str = "%Y%m%d"
    s = ibis.literal(value)
    expr = s.as_date(format_str)

    expr_yaml = compiler.to_yaml(expr)
    yaml_dict = expr_yaml["expression"]
    assert yaml_dict["op"] == "StringToDate"
    assert yaml_dict["arg"]["value"] == value
    assert yaml_dict["format_str"]["value"] == format_str

    roundtrip = compiler.from_yaml(expr_yaml)
    assert roundtrip.equals(expr)


def test_string_length(compiler):
    s = ibis.literal("hello")
    expr = s.length()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "StringLength"
    assert expression["arg"]["value"] == "hello"
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Int32",
        "nullable": {"op": "bool", "value": True},
    }


def test_string_substring(compiler):
    s = ibis.literal("hello world")
    expr = s.substr(0, 5)
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "Substring"
    assert expression["arg"]["value"] == "hello world"
    assert expression["start"]["value"] == 0
    assert expression["length"]["value"] == 5


def test_string_startswith(compiler):
    t = ibis.table({"id": "int32", "name": "string"}, name="T")
    q = t.filter(t.name.startswith("X"))

    yaml_dict = compiler.to_yaml(q)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(q)


def test_string_endswith(compiler):
    t = ibis.table({"id": "int32", "name": "string"}, name="T")
    q = t.filter(t.name.endswith("X"))

    yaml_dict = compiler.to_yaml(q)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(q)


@pytest.mark.parametrize(
    "fun",
    [
        param(
            lambda t: t.string_col.contains("6"),
            id="contains",
        ),
        param(
            lambda t: t.string_col.rlike("|".join(map(str, range(10)))),
            id="rlike",
        ),
        param(
            lambda t: ("a" + t.string_col + "a").re_search(r"\d+"),
            id="re_search_substring",
        ),
        param(
            lambda t: t.string_col.re_search(r"\d+"),
            id="re_search",
        ),
        param(
            lambda t: t.string_col.re_search(r"[[:digit:]]+"),
            id="re_search_posix",
        ),
        param(
            lambda t: t.string_col.re_extract(r"([[:digit:]]+)", 1),
            id="re_extract_posix",
        ),
        param(
            lambda t: (t.string_col + "1").re_extract(r"\d(\d+)", 0),
            id="re_extract_whole_group",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 1),
            id="re_extract_group_1",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 2),
            id="re_extract_group_2",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 3),
            id="re_extract_group_3",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"^(\d+)", 1),
            id="re_extract_group_at_beginning",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)$", 1),
            id="re_extract_group_at_end",
        ),
        param(
            lambda t: t.string_col.re_replace(r"[[:digit:]]+", "a"),
            id="re_replace_posix",
        ),
        param(
            lambda t: t.string_col.re_replace(r"\d+", "a"),
            id="re_replace",
        ),
        param(
            lambda t: t.string_col.repeat(2),
            id="repeat_method",
        ),
        param(
            lambda t: 2 * t.string_col,
            id="repeat_left",
        ),
        param(
            lambda t: t.string_col * 2,
            id="repeat_right",
        ),
        param(
            lambda t: t.string_col.translate("01", "ab"),
            id="translate",
        ),
        param(
            lambda t: t.string_col.find("a"),
            id="find",
        ),
        param(
            lambda t: t.date_string_col.find("13", 3),
            id="find_start",
        ),
        param(
            lambda t: t.string_col.lpad(10, "a"),
            id="lpad",
        ),
        param(
            lambda t: t.string_col.rpad(10, "a"),
            id="rpad",
        ),
        param(
            lambda t: t.string_col.lower(),
            id="lower",
        ),
        param(
            lambda t: t.string_col.upper(),
            id="upper",
        ),
        param(
            lambda t: t.string_col.reverse(),
            id="reverse",
        ),
        param(
            lambda t: t.string_col.ascii_str(),
            id="ascii_str",
        ),
        param(
            lambda t: t.string_col.length(),
            id="length",
        ),
        param(
            lambda t: t.int_col.cases(
                *[(1, "abcd"), (2, "ABCD")], else_="dabc"
            ).startswith("abc"),
            id="startswith",
        ),
        param(
            lambda t: t.date_string_col.startswith("2010-01"),
            id="startswith-simple",
        ),
        param(
            lambda t: t.string_col.strip(),
            id="strip",
        ),
        param(
            lambda t: t.string_col.lstrip(),
            id="lstrip",
        ),
        param(
            lambda t: t.string_col.rstrip(),
            id="rstrip",
        ),
        param(
            lambda t: t.string_col.capitalize(),
            id="capitalize",
        ),
        param(
            lambda t: t.date_string_col.substr(2, 3),
            id="substr",
        ),
        param(
            lambda t: t.date_string_col.substr(2),
            id="substr-start-only",
        ),
        param(
            lambda t: t.date_string_col.left(2),
            id="left",
        ),
        param(
            lambda t: t.date_string_col.right(2),
            id="right",
        ),
        param(
            lambda t: t.date_string_col[1:3],
            id="slice",
        ),
        param(
            lambda t: t.date_string_col[-2],
            id="negative-index",
        ),
        param(
            lambda t: t.date_string_col[: t.date_string_col.length()],
            id="expr_slice_end",
        ),
        param(
            lambda t: t.date_string_col[:],
            id="expr_empty_slice",
        ),
        param(
            lambda t: t.date_string_col[
                t.date_string_col.length() - 2 : t.date_string_col.length() - 1
            ],
            id="expr_slice_begin_end",
        ),
        param(
            lambda t: xo.literal("-").join(["a", t.string_col, "c"]),
            id="join",
        ),
        param(
            lambda t: t.string_col + t.date_string_col,
            id="concat_columns",
        ),
        param(
            lambda t: t.string_col + "a",
            id="concat_column_scalar",
        ),
        param(
            lambda t: "a" + t.string_col,
            id="concat_scalar_column",
        ),
        param(
            lambda t: t.string_col.replace("1", "42"),
            id="replace",
        ),
    ],
)
def test_string(alltypes, compiler, fun):
    expr = fun(alltypes).name("tmp")
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



from __future__ import annotations

import functools
import itertools
import textwrap
import types
from collections.abc import Mapping, Sequence
from typing import Optional

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.graph import Node


_infix_ops = {
    # comparison operations
    ops.Equals: "==",
    ops.IdenticalTo: "===",
    ops.NotEquals: "!=",
    ops.Less: "<",
    ops.LessEqual: "<=",
    ops.Greater: ">",
    ops.GreaterEqual: ">=",
    # arithmetic operations
    ops.Add: "+",
    ops.Subtract: "-",
    ops.Multiply: "*",
    ops.Divide: "/",
    ops.FloorDivide: "//",
    ops.Modulus: "%",
    ops.Power: "**",
    # temporal operations
    ops.DateAdd: "+",
    ops.DateSub: "-",
    ops.DateDiff: "-",
    ops.TimeAdd: "+",
    ops.TimeSub: "-",
    ops.TimeDiff: "-",
    ops.TimestampAdd: "+",
    ops.TimestampSub: "-",
    ops.TimestampDiff: "-",
    ops.IntervalAdd: "+",
    ops.IntervalSubtract: "-",
    ops.IntervalMultiply: "*",
    ops.IntervalFloorDivide: "//",
    # boolean operators
    ops.And: "&",
    ops.Or: "|",
    ops.Xor: "^",
}


def type_info(datatype) -> str:
    """Format `datatype` for display next to a column."""
    return f"  # {datatype}" * ibis.options.repr.show_types


def truncate(pieces: Sequence[str], limit: int) -> list[str]:
    if limit < 1:
        raise ValueError("limit must be >= 1")
    elif limit == 1:
        return pieces[-1:]
    elif limit >= len(pieces):
        return pieces

    first_n = limit // 2
    last_m = limit - first_n
    first, last = pieces[:first_n], pieces[-last_m:]

    maxlen = max(*map(len, first), *map(len, last))
    ellipsis = util.VERTICAL_ELLIPSIS.center(maxlen)

    return [*first, ellipsis, *last]


def render(obj, indent_level=0, limit_items=None, key_separator=":"):
    if isinstance(obj, str):
        result = obj
    elif isinstance(obj, Mapping):
        rendered = {f"{k}{key_separator}": render(v) for k, v in obj.items() if v}
        if not rendered:
            return ""
        maxlen = max(map(len, rendered.keys()))
        lines = [f"{k:<{maxlen}} {v}" for k, v in rendered.items()]
        if limit_items is not None:
            lines = truncate(lines, limit_items)
        result = "\n".join(lines)
    elif isinstance(obj, Sequence):
        lines = tuple(render(item) for item in obj)
        if limit_items is not None:
            lines = truncate(lines, limit_items)
        result = "\n".join(lines)
    else:
        result = str(obj)

    return util.indent(result, spaces=indent_level * 2)


def render_fields(fields, indent_level=0, limit_items=None):
    rendered = {k: render(v, 1) for k, v in fields.items() if v}
    lines = [f"{k}:\n{v}" for k, v in rendered.items()]
    if limit_items is not None:
        lines = truncate(lines, limit_items)
    result = "\n".join(lines)
    return util.indent(result, spaces=indent_level * 2)


def render_schema(schema, indent_level=0, limit_items=None):
    if not len(schema):
        return util.indent("<empty schema>", spaces=indent_level * 2)
    if limit_items is None:
        limit_items = ibis.options.repr.table_columns
    return render(schema, indent_level, limit_items, key_separator="")


def inline(obj):
    if isinstance(obj, Mapping):
        fields = ", ".join(f"{k!r}: {inline(v)}" for k, v in obj.items())
        return f"{{{fields}}}"
    elif util.is_iterable(obj):
        elems = ", ".join(inline(item) for item in obj)
        return f"[{elems}]"
    elif isinstance(obj, types.FunctionType):
        return obj.__name__
    elif isinstance(obj, dt.DataType):
        return str(obj)
    else:
        return repr(obj)


def inline_args(fields, prefer_positional=False):
    fields = {k: inline(v) for k, v in fields.items() if v}

    if fields and prefer_positional:
        first, *rest = fields.keys()
        if not rest:
            return fields[first]
        elif first in {"arg", "expr"}:
            first = fields[first]
            rest = (f"{k}={fields[k]}" for k in rest)
            return ", ".join((first, *rest))

    return ", ".join(f"{k}={v}" for k, v in fields.items())


class Rendered(str):
    def __repr__(self):
        return self


@public
def pretty(node: Node, scope: Optional[dict[str, Node]] = None) -> str:
    """Pretty print an expression.

    Parameters
    ----------
    node
        The graph node to pretty print.
    scope
        A dictionary of expression to name mappings used to intermediate
        assignments. If not provided aliases will be generated for each
        relation.

    Returns
    -------
    str
        A pretty printed representation of the expression.
    """
    if not isinstance(node, Node):
        raise TypeError(f"Expected a graph node, got {type(node)}")

    ref_cnt = itertools.count()
    refs, results = _pretty_by_node(node, scope, ref_cnt)
    out = []
    for ref, rendered in refs.items():
        if ref is not node:
            out.append(f"{results[ref]} := {rendered}")

    res = refs.get(node, results[node])
    if isinstance(node, ops.Literal):
        out.append(res)
    elif isinstance(node, ops.Value):
        out.append(f"{node.name}: {res}{type_info(node.dtype)}")
    else:
        out.append(res)

    return "\n\n".join(out)


def _pretty_by_node(node, scope, ref_cnt, refs=None):
    from xorq.expr.relations import CachedNode, RemoteTable

    if refs is None:
        refs = {}

    variables = {v.op(): k for k, v in (scope or {}).items()}

    results = {}

    def mapper(op, _, **kwargs):
        if isinstance(op, (RemoteTable, CachedNode)):
            attr = "remote_expr" if isinstance(op, RemoteTable) else "parent"
            node = getattr(op, attr).op()
            _refs, _results = _pretty_by_node(node, scope, ref_cnt, refs=refs)
            results.update(_results)
            refs.update(_refs)
            kwargs[attr] = _results.get(node)

        result = fmt(op, **kwargs)
        if var := variables.get(op):
            refs[op] = result
            result = var
        elif isinstance(op, ops.Relation) and not isinstance(op, ops.JoinReference):
            refs[op] = result
            result = f"r{next(ref_cnt)}"
        return Rendered(result)

    results.update(node.map(mapper))
    return refs, results


@functools.singledispatch
def fmt(op, **kwargs):
    top = f"{op.__class__.__name__}\n"
    return top + render_fields(kwargs, 1)


@fmt.register(ops.Relation)
def _relation(op, parent=None, **kwargs):
    if parent is None:
        top = f"{op.__class__.__name__}\n"
    else:
        top = f"{op.__class__.__name__}[{parent}]\n"
    kwargs["schema"] = render_schema(op.schema)
    return top + render_fields(kwargs, 1)


@fmt.register(ops.PhysicalTable)
def _physical_table(op, name, **kwargs):
    schema = render_schema(op.schema, indent_level=1)
    return f"{op.__class__.__name__}: {name}\n{schema}"


@fmt.register(ops.UnboundTable)
@fmt.register(ops.DatabaseTable)
def _unbound_table(op, name, **kwargs):
    schema = render_schema(op.schema, indent_level=1)
    name = ".".join(filter(None, op.namespace.args + (name,)))
    return f"{op.__class__.__name__}: {name}\n{schema}"


@fmt.register(ops.InMemoryTable)
def _in_memory_table(op, data, **kwargs):
    import rich.pretty

    name = f"{op.__class__.__name__}\n"
    data = rich.pretty.pretty_repr(op.data, max_length=ibis.options.repr.table_columns)
    return name + render_fields({"data": data}, 1)


@fmt.register(ops.SQLQueryResult)
@fmt.register(ops.SQLStringView)
def _sql_query_result(op, query, **kwargs):
    clsname = op.__class__.__name__

    if isinstance(op, ops.SQLStringView):
        child = kwargs["child"]
        top = f"{clsname}[{child}]\n"
    else:
        top = f"{clsname}\n"

    query = textwrap.shorten(
        query,
        width=ibis.options.repr.query_text_length,
        placeholder=f" {util.HORIZONTAL_ELLIPSIS}",
    )
    schema = render_schema(op.schema)
    return top + render_fields({"query": query, "schema": schema}, 1)


@fmt.register(ops.FillNull)
@fmt.register(ops.DropNull)
def _fill_null(op, parent, **kwargs):
    name = f"{op.__class__.__name__}[{parent}]\n"
    return name + render_fields(kwargs, 1)


@fmt.register(ops.Aggregate)
def _aggregate(op, parent, **kwargs):
    name = f"{op.__class__.__name__}[{parent}]\n"
    return name + render_fields(kwargs, 1)


@fmt.register(ops.Project)
def _project(op, parent, values):
    name = f"{op.__class__.__name__}[{parent}]\n"

    fields = {}
    for k, v in values.items():
        node = op.values[k]
        fields[f"{k}:"] = f"{v}{type_info(node.dtype)}"

    return name + render_schema(fields, 1)


@fmt.register(ops.DummyTable)
def _dummy_table(op, values):
    name = op.__class__.__name__ + "\n"

    fields = {}
    for k, v in values.items():
        node = op.values[k]
        fields[f"{k}:"] = f"{v}{type_info(node.dtype)}"

    return name + render_schema(fields, 1)


@fmt.register(ops.Filter)
def _project(op, parent, predicates):
    name = f"{op.__class__.__name__}[{parent}]\n"
    return name + render(predicates, 1)


@fmt.register(ops.Sort)
def _sort(op, parent, keys):
    name = f"{op.__class__.__name__}[{parent}]\n"
    return name + render(keys, 1)


@fmt.register(ops.Set)
def _set_op(op, left, right, distinct):
    args = [str(left), str(right)]
    if op.distinct is not None:
        args.append(f"distinct={distinct}")
    return f"{op.__class__.__name__}[{', '.join(args)}]"


@fmt.register(ops.JoinChain)
def _join(op, left, right, predicates, **kwargs):
    args = [str(left), str(right)]
    name = f"{op.__class__.__name__}[{', '.join(args)}]"

    if len(predicates) == 1:
        # if only one, put it directly after the join on the same line
        top = f"{name} {predicates[0]}"
        fields = kwargs
    else:
        top = f"{name}"
        fields = {"predicates": predicates, **kwargs}

    fields = render_fields(fields, 1)
    return f"{top}\n{fields}" if fields else top


@fmt.register(ops.JoinLink)
def _join(op, how, table, predicates):
    args = [str(how), str(table)]
    name = f"{op.__class__.__name__}[{', '.join(args)}]"
    return f"{name}\n{render(predicates, 1)}"


@fmt.register(ops.JoinChain)
def _join_project(op, first, rest, **kwargs):
    name = f"{op.__class__.__name__}[{first}]\n"
    return name + render(rest, 1) + "\n" + render_fields(kwargs, 1)


@fmt.register(ops.Limit)
@fmt.register(ops.Sample)
def _limit(op, parent, **kwargs):
    params = inline_args(kwargs)
    return f"{op.__class__.__name__}[{parent}, {params}]"


@fmt.register(ops.SelfReference)
@fmt.register(ops.Distinct)
def _self_reference(op, parent, **kwargs):
    return f"{op.__class__.__name__}[{parent}]"


@fmt.register(ops.JoinReference)
def _join_reference(op, parent, **kwargs):
    return parent


@fmt.register(ops.Literal)
def _literal(op, value, **kwargs):
    if op.dtype.is_interval():
        return f"{value!r} {op.dtype.unit.short}"
    elif op.dtype.is_array():
        return f"{list(value)!r}"
    else:
        return f"{value!r}"


@fmt.register(ops.Field)
def _relation_field(op, rel, name):
    if name.isidentifier():
        return f"{rel}.{name}"
    else:
        return f"{rel}[{name!r}]"


@fmt.register(ops.Value)
def _value(op, **kwargs):
    fields = inline_args(kwargs, prefer_positional=True)
    return f"{op.__class__.__name__}({fields})"


@fmt.register(ops.Alias)
def _alias(op, arg, name):
    return arg


@fmt.register(ops.Binary)
def _binary(op, left, right):
    try:
        symbol = _infix_ops[op.__class__]
    except KeyError:
        return f"{op.__class__.__name__}({left}, {right})"
    else:
        return f"{left} {symbol} {right}"


@fmt.register(ops.ScalarParameter)
def _scalar_parameter(op, dtype, **kwargs):
    return f"$({dtype})"


@fmt.register(ops.SortKey)
def _sort_key(op, expr, **kwargs):
    return f"{'asc' if op.ascending else 'desc'} {expr}"


@fmt.register(ops.GeoSpatialBinOp)
def _geo_bin_op(op, left, right, **kwargs):
    fields = [left, right, inline_args(kwargs)]
    args = ", ".join(f"{field}" for field in fields if field)
    return f"{op.__class__.__name__}({args})"



import functools
import operator
import types

import dask.base
import toolz

from xorq.common.utils.dask_normalize._ctypes import (
    get_ctypes_field,
)
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    normalize_seq_with_caller,
)
from xorq.common.utils.inspect_utils import (
    get_partial_arguments,
)
from xorq.common.utils.toolz_utils import (
    curry as xo_curry,
)


CODE_ATTRS = (
    "co_argcount",
    "co_cellvars",
    "co_code",
    "co_consts",
    # co_flags | NESTED?
    "co_flags",
    "co_freevars",
    "co_kwonlyargcount",
    "co_name",
    "co_names",
    "co_nlocals",
    "co_stacksize",
    "co_varnames",
    # 'co_lnotab', 'co_filename', 'co_firstlineno',
)
FUNCTION_ATTRS = (
    "__class__",
    "__closure__",
    "__code__",
    "__defaults__",
    "__dict__",
    "__kwdefaults__",
    "__module__",
    "__name__",
    "__qualname__",
)


@toolz.curry
def normalize_by_attrs(attrs, obj):
    objs = tuple(getattr(obj, attr, None) for attr in attrs)
    return normalize_seq_with_caller(*objs)


@dask.base.normalize_token.register(
    (
        types.FunctionType,
        types.MethodType,
        functools._lru_cache_wrapper,
        # HAK: add classmethod:
        #      NOTE: we are *IN*-sensitive to class definition changes
        classmethod,
    )
)
def normalize_function(function):
    def unwrap(obj, attr_name):
        while hasattr(obj, attr_name):
            obj = getattr(obj, attr_name)
        return obj

    function = unwrap(function, "__wrapped__")
    normalized = normalize_by_attrs(FUNCTION_ATTRS, function)
    return normalized


normalize_code = normalize_by_attrs(CODE_ATTRS)
dask.base.normalize_token.register(types.CodeType, normalize_code)
dask.base.normalize_token.register(property, normalize_code)


@dask.base.normalize_token.register(toolz.functoolz.Compose)
def normalize_toolz_compose(composed):
    return normalize_seq_with_caller(
        toolz.functoolz.Compose,
        composed.first,
        composed.funcs,
    )


@dask.base.normalize_token.register((toolz.curry, xo_curry))
def normalize_toolz_curry(curried):
    partial_arguments = get_partial_arguments(
        curried.func, *curried.args, **curried.keywords
    )
    objs = sum(
        map(
            dask.base.normalize_token,
            (
                curried.func,
                # FIXME: register dict normalization to fix order?
                sorted(partial_arguments.items()),
            ),
        ),
        start=(),
    )
    return objs


def make_cell_typ():
    # https://stackoverflow.com/a/23830790
    def outer(x):
        def inner(y):
            return x * y

        return inner

    inner = outer(1)
    typ = type(inner.__closure__[0])
    return typ


@dask.base.normalize_token.register(make_cell_typ())
def normalize_cell(cell):
    return dask.base.normalize_token(cell.cell_contents)


@dask.base.normalize_token.register(toolz.functoolz.excepts)
def normalize_excepts(f):
    return normalize_seq_with_caller(
        f.exc,
        f.func,
        # FIXME: figure out how to include handler
        # f.handler,
    )


@dask.base.normalize_token.register(operator.methodcaller)
def normalize_operator_methodcaller(obj):
    fields = ("name", "args", "kwargs")
    gen = (get_ctypes_field(fields, field, obj) for field in fields)
    return normalize_seq_with_caller(*gen)



"""Cross-validation for xorq ML pipelines."""

from random import Random

import cloudpickle
import dask
import numpy as np
import pyarrow as pa
import toolz
from attr import field, frozen
from attr.validators import deep_iterable, instance_of

import xorq.expr.datatypes as dt
from xorq.expr.udf import pyarrow_udwf
from xorq.vendor import ibis
from xorq.vendor.ibis.expr.api import literal
from xorq.vendor.ibis.expr.types import Expr


def make_deterministic_sort_key(expr, random_seed=None):
    """Make a deterministic sort key for an ibis table.

    Concatenates every column as a string (comma-separated), appends a
    deterministic random string derived from *random_seed*, and hashes the
    result.  Two tables with the same data and the same seed will always
    produce the same per-row hash within a given backend.

    Use this to impose a stable row order before any operation that depends
    on positional indexing (e.g. sklearn splitters).  See
    :func:`apply_deterministic_sort` for a convenience wrapper that
    returns a sorted ibis expression directly.

    .. note::

       The hash function is backend-dependent, so the same data may
       produce different sort orders on different backends.  Within a
       single backend, the ordering is fully deterministic.

    Parameters
    ----------
    expr : ir.Table
        The input table.
    random_seed : int or None
        Seed for reproducibility.  ``None`` uses a fixed default (0).

    Returns
    -------
    ir.IntegerColumn
        Hash column suitable for use in ``order_by``.
    """
    if random_seed is None:
        random_seed = 0
    random_str = str(Random(random_seed).getrandbits(256))
    tmp_name = "_sort_" + dask.base.tokenize(random_str)[:8]
    comb_key = literal(",").join(expr[col].cast("str") for col in expr.columns)
    return comb_key.concat(random_str).hash().name(tmp_name)


def apply_deterministic_sort(expr, random_seed=None):
    """Sort an ibis table by a deterministic hash.

    Convenience wrapper around :func:`make_deterministic_sort_key` that
    mutates, sorts, and drops the temporary column, returning a deferred
    ibis expression.  Call ``.execute()`` to materialize.

    Parameters
    ----------
    expr : ir.Table
        The input table.
    random_seed : int or None
        Seed for reproducibility.  ``None`` uses a fixed default (0).

    Returns
    -------
    ir.Table
        The input table sorted by a deterministic hash.
    """
    key = make_deterministic_sort_key(expr, random_seed=random_seed)
    col = key.get_name()
    return expr.mutate(key).order_by(col).drop(col)


def _make_folds_from_int(expr, cv, random_seed):
    """Build a fold expression from k equal-sized hash-based folds.

    Returns the input table with ``fold_0``, ``fold_1``, ..., ``fold_{k-1}``
    columns appended (int8, 0=unused, 1=train, 2=test).

    Parameters
    ----------
    expr : ir.Table
        The input table to split.
    cv : int
        Number of folds.
    random_seed : int or None
        Random seed for reproducibility.

    Returns
    -------
    ir.Table
        The input table with ``fold_0``, ..., ``fold_{k-1}`` columns
        appended (int8, 0=unused, 1=train, 2=test).
    """
    from xorq.expr.ml.split_lib import calc_split_column

    test_sizes = tuple(1.0 / cv for _ in range(cv))
    split_col_name = "__cv_split__"
    split_col = calc_split_column(
        expr,
        unique_key=expr.columns,
        test_sizes=test_sizes,
        num_buckets=10000,
        random_seed=random_seed,
        name=split_col_name,
    )
    base_expr = expr.mutate(split_col)
    # Encode: 2=test (row belongs to fold_i), 1=train (row belongs to another fold)
    fold_columns = [
        ((base_expr[split_col_name] == fold_i).cast("int8") + 1).name(f"fold_{fold_i}")
        for fold_i in range(cv)
    ]
    return base_expr.mutate(*fold_columns).drop(split_col_name)


def _make_fold_udwf(cv, fold_index, features, target, expr):
    """Create a UDWF that marks rows for a given fold.

    The UDWF receives feature and target columns for all rows, runs the sklearn
    splitter to determine fold assignments, and returns a per-row int8 column
    where 0 = unused, 1 = train, 2 = test.

    Parameters
    ----------
    cv : sklearn splitter
        An sklearn cross-validation splitter with a .split() method.
    fold_index : int
        Which fold (0-based) this UDWF assigns test membership for.
    features : tuple of str
        Feature column names.
    target : str
        Target column name.
    expr : ir.Table
        The input table (used to infer column types for the UDWF schema).

    Returns
    -------
    callable
        A UDWF constructor with an .on_expr() method.
    """
    # Build the schema from the actual expression types (features + target)
    # We need to serialize cv because pyarrow_udwf stores config in FrozenDict
    # and sklearn splitters aren't hashable.  We pickle them and unpickle
    # inside the UDWF.
    cv_bytes = cloudpickle.dumps(cv)

    table_schema = expr.schema()
    schema_cols = tuple((col, table_schema[col]) for col in (*features, target))

    @pyarrow_udwf(
        schema=ibis.schema(schema_cols),
        return_type=dt.int8,
        cv_bytes=cv_bytes,
        fold_index=fold_index,
        n_features=len(features),
        name=f"cv_fold_{fold_index}",
    )
    def assign_fold(self, values, num_rows):
        import cloudpickle as cp

        cv_splitter = cp.loads(self.cv_bytes)
        n_feat = self.n_features
        fold_i = self.fold_index

        # values[:n_feat] are feature columns, values[n_feat] is target
        X = np.column_stack([v.to_numpy(zero_copy_only=False) for v in values[:n_feat]])
        y = values[n_feat].to_numpy(zero_copy_only=False)

        # Run the splitter: 0=unused, 1=train, 2=test
        train_idx, test_idx = next(
            (train, test)
            for i, (train, test) in enumerate(cv_splitter.split(X, y))
            if i == fold_i
        )
        result = np.zeros(num_rows, dtype=np.int8)
        result[train_idx] = 1
        result[test_idx] = 2

        return pa.array(result, type=pa.int8())

    return assign_fold


def _resolve_order_by(expr, order_by, random_seed):
    """Resolve *order_by* into a sorted expression.

    Parameters
    ----------
    expr : ir.Table
        The input table.
    order_by : str, tuple of str, or None
        Column(s) to sort by.  ``None`` falls back to a deterministic
        hash derived from *random_seed*.
    random_seed : int or None
        Seed for the deterministic row-ordering hash (used only when
        *order_by* is ``None``).

    Returns
    -------
    ir.Table
        The sorted expression.
    """
    match order_by:
        case str() | tuple() | list() as key:
            return expr.order_by(key)
        case None:
            return apply_deterministic_sort(expr, random_seed)
        case _:
            raise TypeError(
                f"order_by must be a str, tuple of str, or None, "
                f"got {type(order_by).__name__}"
            )


def _make_folds_from_sklearn(
    expr, cv, features, target, random_seed=None, order_by=None
):
    """Build a fold expression from an sklearn splitter using deferred UDWFs.

    Returns the input table with ``fold_0``, ``fold_1``, ..., ``fold_{k-1}``
    columns appended (int8, 0=unused, 1=train, 2=test).

    Row ordering for the UDWF is determined by:

    1. If *order_by* is provided, rows are sorted by those column(s).
    2. Otherwise, rows are sorted by a deterministic seeded hash so
       the UDWF always sees the same order regardless of backend scan
       order.

    Parameters
    ----------
    expr : ir.Table
        The input table.
    cv : sklearn splitter
        An sklearn cross-validation splitter with a .split() method.
    features : tuple of str
        Feature column names (used as X for stratified splitters).
    target : str
        Target column name (used as y for stratified splitters).
    random_seed : int or None
        Seed for the deterministic row-ordering hash.
    order_by : str, tuple of str, or None
        Column(s) to sort by before folding.  Overrides the default
        hash-based sort.  Useful for ``TimeSeriesSplit`` where rows
        must be in temporal order.

    Returns
    -------
    ir.Table
        The input table with ``fold_0``, ..., ``fold_{k-1}`` columns
        appended (int8, 0=unused, 1=train, 2=test).
    """
    n_splits = cv.get_n_splits()
    window = ibis.window()
    sorted_expr = _resolve_order_by(expr, order_by, random_seed)

    fold_columns = [
        _make_fold_udwf(cv, i, features, target, sorted_expr)
        .on_expr(sorted_expr)
        .over(window)
        .name(f"fold_{i}")
        for i in range(n_splits)
    ]
    return sorted_expr.mutate(*fold_columns)


def _fold_pairs_from_fold_expr(fold_expr, n_splits):
    """Extract (train, test) pairs from a fold expression.

    Parameters
    ----------
    fold_expr : ir.Table
        Table with ``fold_0``, ..., ``fold_{k-1}`` columns
        (int8, 0=unused, 1=train, 2=test).
    n_splits : int
        Number of folds.

    Returns
    -------
    tuple of (ir.Table, ir.Table)
        One (train, test) pair per fold.  Fold columns are dropped from
        the returned tables.
    """
    fold_cols = [f"fold_{i}" for i in range(n_splits)]
    return tuple(
        (
            fold_expr.filter(fold_expr[f"fold_{i}"] == 1).drop(*fold_cols),
            fold_expr.filter(fold_expr[f"fold_{i}"] == 2).drop(*fold_cols),
        )
        for i in range(n_splits)
    )


@frozen
class CrossValScore:
    """Deferred cross-validation scores.

    Holds a tuple of per-fold ibis score expressions and a ``fold_expr`` —
    a single ibis table with the original columns plus ``fold_0``,
    ``fold_1``, ..., ``fold_k`` columns (int8, 0 = unused, 1 = train, 2 = test).

    Call ``.execute()`` to materialize the scores, or inspect
    ``fold_expr`` to visualize / debug fold assignments.
    """

    score_exprs = field(
        validator=deep_iterable(instance_of(Expr), instance_of(tuple)),
        converter=tuple,
    )
    fold_expr = field(validator=instance_of(Expr))

    def execute(self):
        """Materialize all per-fold scores into a numpy array.

        Returns
        -------
        numpy.ndarray
            Array of scores, one per fold.
        """
        return np.array(tuple(expr.execute() for expr in self.score_exprs))

    def __len__(self):
        return len(self.score_exprs)


@toolz.curry
def deferred_cross_val_score(
    pipeline,
    expr,
    features,
    target,
    cv=5,
    scoring=None,
    random_seed=None,
    order_by=None,
):
    """Evaluate a pipeline using cross-validation, returning deferred per-fold scores.

    This is the xorq equivalent of sklearn's cross_val_score. The pipeline is
    fit and scored on each fold independently. Returns a CrossValScore holding
    deferred ibis expressions — call .execute() to materialize the scores.

    Parameters
    ----------
    pipeline : Pipeline
        An unfitted xorq Pipeline (will be fit on each fold's training data).
    expr : ir.Table
        The input ibis table expression containing features and target.
    features : tuple of str
        Feature column names.
    target : str
        Target column name.
    cv : int or sklearn splitter
        Cross-validation strategy. If int, uses train_test_splits to create
        that many equal-sized folds. If an sklearn splitter object (e.g.
        KFold, StratifiedKFold), uses its .split() method to generate
        train/test indices. Default is 5.
    scoring : str, callable, _BaseScorer, Scorer, optional
        Scorer specification passed to FittedPipeline.score_expr(). If None,
        uses the model's default scorer (accuracy for classifiers, r2 for
        regressors). Default is None.
    random_seed : int, optional
        Random seed for reproducibility.  When cv is an int, controls
        hash-based fold partitioning.  When cv is an sklearn splitter,
        controls the deterministic row ordering used to guarantee that
        the UDWF sees rows in a stable order.  To reproduce the same
        fold assignments with standalone sklearn, sort the pandas
        DataFrame with :func:`apply_deterministic_sort` using the same
        seed before calling ``cross_val_score``.  Default is None.
    order_by : str, tuple of str, optional
        Column(s) to sort by before folding.  Overrides the default
        hash-based sort.  Required for ``TimeSeriesSplit`` to specify
        the temporal ordering column.  Default is None.

    Returns
    -------
    CrossValScore
        Deferred cross-validation result. Call .execute() to get numpy.ndarray.

    Examples
    --------
    >>> from sklearn.pipeline import Pipeline as SklearnPipeline
    >>> from sklearn.preprocessing import StandardScaler
    >>> from sklearn.linear_model import LogisticRegression
    >>> import xorq.api as xo
    >>> from xorq.expr.ml.pipeline_lib import Pipeline
    >>> from xorq.expr.ml.cross_validation import deferred_cross_val_score
    >>>
    >>> t = xo.memtable({"x1": range(100), "x2": range(100), "y": [0, 1] * 50})
    >>> pipe = Pipeline.from_instance(SklearnPipeline([
    ...     ("scaler", StandardScaler()),
    ...     ("clf", LogisticRegression()),
    ... ]))
    >>> cv_scores = deferred_cross_val_score(pipe, t, features=("x1", "x2"), target="y", cv=5)
    >>> scores = cv_scores.execute()  # materializes all folds
    """
    from sklearn.model_selection import TimeSeriesSplit
    from sklearn.model_selection._split import GroupsConsumerMixin

    from xorq.expr.ml.pipeline_lib import Pipeline

    if not isinstance(pipeline, Pipeline):
        raise TypeError(
            f"pipeline must be a Pipeline instance, got {type(pipeline).__name__}"
        )

    features = tuple(features)

    match cv:
        case int():
            fold_expr = _make_folds_from_int(expr, cv, random_seed)
            n_splits = cv
        case TimeSeriesSplit() if order_by is None:
            raise TypeError(
                "TimeSeriesSplit requires order_by to specify the "
                "temporal ordering column(s)."
            )
        case GroupsConsumerMixin():
            raise TypeError(
                f"Group-based splitters are not supported "
                f"(got {type(cv).__name__}). "
                f"Use a non-group splitter such as KFold, "
                f"StratifiedKFold, ShuffleSplit, etc."
            )
        case object(split=_):
            fold_expr = _make_folds_from_sklearn(
                expr,
                cv,
                features,
                target,
                random_seed=random_seed,
                order_by=order_by,
            )
            n_splits = cv.get_n_splits()
        case _:
            raise TypeError(
                f"cv must be an int or an sklearn splitter with a .split() method, "
                f"got {type(cv).__name__}"
            )

    fold_pairs = _fold_pairs_from_fold_expr(fold_expr, n_splits)

    score_exprs = tuple(
        pipeline.fit(train, features=features, target=target).score_expr(
            test, scorer=scoring
        )
        for train, test in fold_pairs
    )

    return CrossValScore(score_exprs=score_exprs, fold_expr=fold_expr)


__all__ = [
    "apply_deterministic_sort",
    "deferred_cross_val_score",
    "make_deterministic_sort_key",
]



from __future__ import annotations

import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.tests.util import assert_frame_equal, assert_series_equal


@pytest.mark.parametrize(
    "text_value",
    [
        param(
            "STRING",
            id="string",
        ),
        param(
            "STRI'NG",
            id="string-quote1",
        ),
        param(
            'STRI"NG',
            id="string-quote2",
        ),
    ],
)
def test_string_literal(con, text_value):
    expr = xo.literal(text_value)
    result = con.execute(expr)
    assert result == text_value


def is_text_type(x):
    return isinstance(x, str)


def test_string_col_is_unicode(alltypes, alltypes_df):
    dtype = alltypes.string_col.type()
    assert dtype == dt.String(nullable=dtype.nullable)
    assert alltypes_df.string_col.map(is_text_type).all()
    result = alltypes.string_col.execute()
    assert result.map(is_text_type).all()


@pytest.mark.parametrize(
    ("result_func", "expected_func"),
    [
        param(
            lambda t: t.string_col.contains("6"),
            lambda t: t.string_col.str.contains("6"),
            id="contains",
        ),
        param(
            lambda t: t.string_col.rlike("|".join(map(str, range(10)))),
            lambda t: t.string_col == t.string_col,
            id="rlike",
        ),
        param(
            lambda t: ("a" + t.string_col + "a").re_search(r"\d+"),
            lambda t: t.string_col.str.contains(r"\d+"),
            id="re_search_substring",
        ),
        param(
            lambda t: t.string_col.re_search(r"\d+"),
            lambda t: t.string_col.str.contains(r"\d+"),
            id="re_search",
        ),
        param(
            lambda t: t.string_col.re_search(r"[[:digit:]]+"),
            lambda t: t.string_col.str.contains(r"\d+"),
            id="re_search_posix",
        ),
        param(
            lambda t: t.string_col.re_extract(r"([[:digit:]]+)", 1),
            lambda t: t.string_col.str.extract(r"(\d+)", expand=False),
            id="re_extract_posix",
        ),
        param(
            lambda t: (t.string_col + "1").re_extract(r"\d(\d+)", 0),
            lambda t: (t.string_col + "1").str.extract(r"(\d+)", expand=False),
            id="re_extract_whole_group",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 1),
            lambda t: t.date_string_col.str.extract(
                r"(\d+)\D(\d+)\D(\d+)", expand=False
            ).iloc[:, 0],
            id="re_extract_group_1",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 2),
            lambda t: t.date_string_col.str.extract(
                r"(\d+)\D(\d+)\D(\d+)", expand=False
            ).iloc[:, 1],
            id="re_extract_group_2",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 3),
            lambda t: t.date_string_col.str.extract(
                r"(\d+)\D(\d+)\D(\d+)", expand=False
            ).iloc[:, 2],
            id="re_extract_group_3",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"^(\d+)", 1),
            lambda t: t.date_string_col.str.extract(r"^(\d+)", expand=False),
            id="re_extract_group_at_beginning",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)$", 1),
            lambda t: t.date_string_col.str.extract(r"(\d+)$", expand=False),
            id="re_extract_group_at_end",
        ),
        param(
            lambda t: t.string_col.re_replace(r"[[:digit:]]+", "a"),
            lambda t: t.string_col.str.replace(r"\d+", "a", regex=True),
            id="re_replace_posix",
        ),
        param(
            lambda t: t.string_col.re_replace(r"\d+", "a"),
            lambda t: t.string_col.str.replace(r"\d+", "a", regex=True),
            id="re_replace",
        ),
        param(
            lambda t: t.string_col.repeat(2),
            lambda t: t.string_col * 2,
            id="repeat_method",
        ),
        param(
            lambda t: 2 * t.string_col,
            lambda t: 2 * t.string_col,
            id="repeat_left",
        ),
        param(
            lambda t: t.string_col * 2,
            lambda t: t.string_col * 2,
            id="repeat_right",
        ),
        param(
            lambda t: t.string_col.translate("01", "ab"),
            lambda t: t.string_col.str.translate(str.maketrans("01", "ab")),
            id="translate",
        ),
        param(
            lambda t: t.string_col.find("a"),
            lambda t: t.string_col.str.find("a"),
            id="find",
        ),
        param(
            lambda t: t.date_string_col.find("13", 3),
            lambda t: t.date_string_col.str.find("13", 3),
            id="find_start",
        ),
        param(
            lambda t: t.string_col.lpad(10, "a"),
            lambda t: t.string_col.str.pad(10, fillchar="a", side="left"),
            id="lpad",
        ),
        param(
            lambda t: t.string_col.rpad(10, "a"),
            lambda t: t.string_col.str.pad(10, fillchar="a", side="right"),
            id="rpad",
        ),
        param(
            lambda t: t.string_col.lower(),
            lambda t: t.string_col.str.lower(),
            id="lower",
        ),
        param(
            lambda t: t.string_col.upper(),
            lambda t: t.string_col.str.upper(),
            id="upper",
        ),
        param(
            lambda t: t.string_col.reverse(),
            lambda t: t.string_col.str[::-1],
            id="reverse",
        ),
        param(
            lambda t: t.string_col.ascii_str(),
            lambda t: t.string_col.map(ord).astype("int32"),
            id="ascii_str",
        ),
        param(
            lambda t: t.string_col.length(),
            lambda t: t.string_col.str.len().astype("int32"),
            id="length",
        ),
        param(
            lambda t: t.int_col.cases(
                *[(1, "abcd"), (2, "ABCD")], else_="dabc"
            ).startswith("abc"),
            lambda t: t.int_col == 1,
            id="startswith",
        ),
        param(
            lambda t: t.date_string_col.startswith("2010-01"),
            lambda t: t.date_string_col.str.startswith("2010-01"),
            id="startswith-simple",
        ),
        param(
            lambda t: t.string_col.strip(),
            lambda t: t.string_col.str.strip(),
            id="strip",
        ),
        param(
            lambda t: t.string_col.lstrip(),
            lambda t: t.string_col.str.lstrip(),
            id="lstrip",
        ),
        param(
            lambda t: t.string_col.rstrip(),
            lambda t: t.string_col.str.rstrip(),
            id="rstrip",
        ),
        param(
            lambda t: t.string_col.capitalize(),
            lambda t: t.string_col.str.capitalize(),
            id="capitalize",
        ),
        param(
            lambda t: t.date_string_col.substr(2, 3),
            lambda t: t.date_string_col.str[2:5],
            id="substr",
        ),
        param(
            lambda t: t.date_string_col.substr(2),
            lambda t: t.date_string_col.str[2:],
            id="substr-start-only",
        ),
        param(
            lambda t: t.date_string_col.left(2),
            lambda t: t.date_string_col.str[:2],
            id="left",
        ),
        param(
            lambda t: t.date_string_col.right(2),
            lambda t: t.date_string_col.str[-2:],
            id="right",
        ),
        param(
            lambda t: t.date_string_col[1:3],
            lambda t: t.date_string_col.str[1:3],
            id="slice",
        ),
        param(
            lambda t: t.date_string_col[-2],
            lambda t: t.date_string_col.str[-2],
            id="negative-index",
        ),
        param(
            lambda t: t.date_string_col[: t.date_string_col.length()],
            lambda t: t.date_string_col,
            id="expr_slice_end",
        ),
        param(
            lambda t: t.date_string_col[:],
            lambda t: t.date_string_col,
            id="expr_empty_slice",
        ),
        param(
            lambda t: t.date_string_col[
                t.date_string_col.length() - 2 : t.date_string_col.length() - 1
            ],
            lambda t: t.date_string_col.str[-2:-1],
            id="expr_slice_begin_end",
        ),
        param(
            lambda t: xo.literal("-").join(["a", t.string_col, "c"]),
            lambda t: "a-" + t.string_col + "-c",
            id="join",
        ),
        param(
            lambda t: t.string_col + t.date_string_col,
            lambda t: t.string_col + t.date_string_col,
            id="concat_columns",
        ),
        param(
            lambda t: t.string_col + "a",
            lambda t: t.string_col + "a",
            id="concat_column_scalar",
        ),
        param(
            lambda t: "a" + t.string_col,
            lambda t: "a" + t.string_col,
            id="concat_scalar_column",
        ),
        param(
            lambda t: t.string_col.replace("1", "42"),
            lambda t: t.string_col.str.replace("1", "42"),
            id="replace",
        ),
    ],
)
def test_string(alltypes, alltypes_df, result_func, expected_func):
    expr = result_func(alltypes).name("tmp")
    result = expr.execute()

    expected = expected_func(alltypes_df)
    assert_series_equal(result, expected)


def test_re_replace_global(con):
    expr = xo.literal("aba").re_replace("a", "c")
    result = con.execute(expr)
    assert result == "cbc"


def test_substr_with_null_values(alltypes, alltypes_df):
    table = alltypes.mutate(
        substr_col_null=xo.cases(
            (alltypes["bool_col"], alltypes["string_col"]), else_=None
        ).substr(0, 2)
    )
    result = table.execute()

    expected = alltypes_df.copy()
    mask = ~expected["bool_col"]
    expected["substr_col_null"] = expected["string_col"]
    expected.loc[mask, "substr_col_null"] = None
    expected["substr_col_null"] = expected["substr_col_null"].str.slice(0, 2)

    assert_frame_equal(result.fillna(pd.NA), expected.fillna(pd.NA))


@pytest.mark.parametrize(
    ("result_func", "expected"),
    [
        param(lambda d: d.protocol(), "http", id="protocol"),
        param(
            lambda d: d.authority(),
            "user:pass@example.com:80",
            id="authority",
        ),
        param(
            lambda d: d.userinfo(),
            "user:pass",
            id="userinfo",
        ),
        param(
            lambda d: d.host(),
            "example.com",
            id="host",
        ),
        param(lambda d: d.path(), "/docs/books/tutorial/index.html", id="path"),
        param(lambda d: d.query(), "name=networking", id="query"),
        param(lambda d: d.query("name"), "networking", id="query-key"),
        param(
            lambda d: d.query(xo.literal("na") + xo.literal("me")),
            "networking",
            id="query-dynamic-key",
        ),
        param(lambda d: d.fragment(), "DOWNLOADING", id="ref"),
    ],
)
def test_parse_url(con, result_func, expected):
    url = "http://user:pass@example.com:80/docs/books/tutorial/index.html?name=networking#DOWNLOADING"
    expr = result_func(xo.literal(url))
    result = con.execute(expr)
    assert result == expected


def test_capitalize(con):
    s = xo.literal("aBc")
    expected = "Abc"
    expr = s.capitalize()
    assert con.execute(expr) == expected


def test_subs_with_re_replace(con):
    expr = xo.literal("hi").re_replace("i", "a").substitute({"d": "b"}, else_="k")
    result = con.execute(expr)
    assert result == "k"


def test_multiple_subs(con):
    m = {"foo": "FOO", "bar": "BAR"}
    expr = xo.literal("foo").substitute(m)
    result = con.execute(expr)
    assert result == "FOO"


@pytest.mark.parametrize(
    "expr",
    [
        param(xo.cases((True, "%")), id="case"),
        param(xo.ifelse(True, "%", xo.null()), id="ifelse"),
    ],
)
def test_no_conditional_percent_escape(con, expr):
    assert con.execute(expr) == "%"


def test_string_length(con):
    t = xo.memtable({"s": ["aaa", "a", "aa"]})
    assert con.execute(t.s.length()).gt(0).all()


def test_hash(con):
    s = xo.literal("aBc")
    expected = 5357040098349975229
    expr = s.hash()
    assert con.execute(expr) == expected


def test_hash_column(alltypes):
    res = alltypes.string_col.hash().name("tmp").execute()
    assert res is not None
